# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import typing as ta


@dc.dataclass(frozen=True)
class ManifestOrigin:
    module: str
    attr: ta.Optional[str]  # None if inline

    file: str
    line: int


@dc.dataclass(frozen=True)
class Manifest(ManifestOrigin):
    value: ta.Any



"""
https://docs.python.org/3/howto/logging.html#configuring-logging
"""
import dataclasses as dc
import typing as ta


##


FilterConfig: ta.TypeAlias = dict[str, ta.Any]
FormatterConfig: ta.TypeAlias = dict[str, ta.Any]
HandlerConfig: ta.TypeAlias = dict[str, ta.Any]
LoggerConfig: ta.TypeAlias = dict[str, ta.Any]


@dc.dataclass()
class DictConfig:
    version: int = 1
    incremental: bool = False
    disable_existing_loggers: bool = False
    filters: dict[str, FilterConfig] = dc.field(default_factory=dict)
    formatters: dict[str, FormatterConfig] = dc.field(default_factory=dict)
    handlers: dict[str, HandlerConfig] = dc.field(default_factory=dict)
    loggers: dict[str, LoggerConfig] = dc.field(default_factory=dict)
    root: LoggerConfig | None = None



import re
import uuid

from ... import check
from ..base import MarshalContext
from ..base import Marshaler
from ..base import TypeMapMarshalerFactory
from ..base import TypeMapUnmarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value


PATTERN = re.compile(r'([0-9A-Fa-f]{8}-([0-9A-Fa-f]{4}-){3}[0-9A-Fa-f]{12})|([0-9A-Fa-f]{32})')


class UuidMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: uuid.UUID) -> Value:
        return str(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> uuid.UUID:
        return uuid.UUID(check.isinstance(v, str).replace('-', ''))


UUID_MARSHALER_UNMARSHALER = UuidMarshalerUnmarshaler()

UUID_MARSHALER_FACTORY = TypeMapMarshalerFactory({uuid.UUID: UUID_MARSHALER_UNMARSHALER})
UUID_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({uuid.UUID: UUID_MARSHALER_UNMARSHALER})



import dataclasses as dc
import typing as ta

from .errors import MalformedNuhError


@dc.dataclass(frozen=True)
class Nuh:
    name: str | None = None
    user: str | None = None
    host: str | None = None

    @property
    def tuple(self) -> tuple[str | None, ...]:
        return (self.name, self.user, self.host)

    def __iter__(self) -> ta.Iterator[str | None]:
        return iter(self.tuple)

    @classmethod
    def parse(cls, inp: str) -> 'Nuh':
        if not inp:
            raise MalformedNuhError

        host: str | None = None
        host_start = inp.find('@')
        if host_start != -1:
            host = inp[host_start + 1:]
            inp = inp[:host_start]

        user_start = inp.find('!')
        user: str | None = None
        if user_start != -1:
            user = inp[user_start + 1:]
            inp = inp[:user_start]

        return cls(
            name=inp or None,
            user=user,
            host=host,
        )

    @property
    def canonical(self) -> str:
        parts = []
        if (n := self.name) is not None:
            parts.append(n)
        if (u := self.user) is not None:
            parts.append(f'!{u}')
        if (h := self.host) is not None:
            parts.append(f'@{h}')
        return ''.join(parts)



"""
TODO:
 - metaclass, to forbid __subclasscheck__ / __instancecheck__? Clash with dc.Meta, don't want that lock-in, not
   necessary for functionality, just a helpful misuse prevention.
"""
import abc
import copy
import dataclasses as dc
import typing as ta

from .. import lang
from ..lite.dataclasses import is_immediate_dataclass
from .impl.api import dataclass


##


class Static(lang.Abstract):
    """
    Dataclass mixin for dataclasses in which all fields have class-level defaults and are not intended for any
    instance-level overrides - effectively making their subclasses equivalent to instances. For dataclasses which make
    sense as singletons (such as project specs or manifests), for which dynamic instantiation is not the usecase, this
    can enable more natural syntax. Inheritance is permitted - equivalent to a `functools.partial` or composing
    `**kwargs`, as long as there is only ever one unambiguous underlying non-static dataclass to be instantiated.
    """

    __static_dataclass_class__: ta.ClassVar[type]
    __static_dataclass_instance__: ta.ClassVar[ta.Any]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        for a in ('__new__', '__init__'):
            if a in cls.__dict__ and not (cls.__dict__[a] == getattr(Static, a)):
                raise TypeError(f'Static base class {cls} must not implement {a}')

        if cls.__init__ is not Static.__init__:
            # This is necessary to make type-checking work (by allowing it to accept zero args). This isn't strictly
            # necessary, but since it's useful to do sometimes it might as well be done everywhere to prevent clashing.
            raise TypeError(f'Static.__init__ should be first in mro of {cls}')

        #

        b_dc_lst = []

        def b_dc_rec(b_cls: type) -> None:  # noqa
            if issubclass(b_cls, Static):
                return
            elif is_immediate_dataclass(b_cls):
                b_dc_lst.append(b_cls)
            elif dc.is_dataclass(b_cls):
                for sb_cls in b_cls.__bases__:
                    b_dc_rec(sb_cls)

        for b_cls in cls.__bases__:
            b_dc_rec(b_cls)

        b_sdc_dc_lst = [
            b_cls.__static_dataclass_class__
            for b_cls in cls.__bases__
            if issubclass(b_cls, Static)
            and b_cls is not Static
        ]

        sdc_dc_set = {*b_dc_lst, *b_sdc_dc_lst}
        if not sdc_dc_set:
            raise TypeError(f'Static dataclass {cls} inherits from no dataclass')

        # Base static dataclasses of various types are allowed as long as there is exactly one final subclass involved.
        # For example, a ModAttrManifest dataclass with an abstract StaticModAttrManifest subclass which sets a default
        # mod_name may be mixed in with a further down use-case specific Manifest subclass.
        sdc_dc_set -= {
            m_cls
            for s_cls in sdc_dc_set
            for m_cls in s_cls.__mro__
            if m_cls is not s_cls
        }

        if len(sdc_dc_set) > 1:
            raise TypeError(f'Static dataclass {cls} inherits from multiple dataclasses: {sdc_dc_set!r}')
        [sdc_cls] = sdc_dc_set

        if '__static_dataclass_class__' in cls.__dict__:
            raise AttributeError
        setattr(cls, '__static_dataclass_class__', sdc_cls)

        #

        expected_fld_order: ta.Sequence[str] | None = None
        is_abstract = lang.is_abstract_class(cls) or abc.ABC in cls.__bases__
        if not is_abstract:
            if is_immediate_dataclass(cls):
                raise TypeError(cls)

            flds_dct = {}
            for b_cls in cls.__mro__:
                if not dc.is_dataclass(b_cls):
                    continue
                b_flds = dc.fields(b_cls)  # noqa
                for fld in b_flds:
                    if fld.name in flds_dct:
                        continue
                    flds_dct[fld.name] = fld

            # Annotations are the authoritative source of field order.
            new_anns = {}

            for fld in flds_dct.values():
                new_fld = copy.copy(fld)

                try:
                    v = cls.__dict__[fld.name]
                except KeyError:
                    if fld.default is dc.MISSING and fld.default_factory is dc.MISSING:
                        raise TypeError(f'Field {fld.name!r} of class {cls} is not set and has no default') from None
                else:
                    if isinstance(v, dc.Field):
                        raise TypeError(f'Static dataclass {cls} may not introduce new fields: {fld.name}: {v}')

                    new_fld.default = dc.MISSING
                    # Use a default_factory to allow unsafe (mutable) values.
                    new_fld.default_factory = (lambda v2: lambda: v2)(v)  # noqa

                setattr(cls, fld.name, new_fld)
                new_anns[fld.name] = fld.type

            # Non-abstract static dataclasses may not introduce new fields.
            expected_fld_order = list(flds_dct)

            new_anns.update({
                k: v
                for k, v in getattr(cls, '__annotations__', {}).items()
                if k not in new_anns
            })

            cls.__annotations__ = new_anns

        else:
            for b_cls in cls.__bases__:
                if hasattr(b_cls, '__static_dataclass_instance__'):
                    raise TypeError(
                        f'Abstract static dataclass {cls} may not inherit from non-abstract static dataclass {b_cls}',
                    )

        # Explicitly forbid dc transforms that rebuild the class, such as slots.
        if (dc_cls := dataclass(cls, frozen=True)) is not cls:
            raise TypeError(dc_cls)

        dc_flds = dc.fields(cls)  # type: ignore[arg-type]  # noqa

        if expected_fld_order is not None:
            dc_fld_order = [f.name for f in dc_flds]
            if dc_fld_order != expected_fld_order:
                raise TypeError(
                    f'Static dataclass field order {dc_fld_order!r} != expected field order {expected_fld_order!r}',
                )

        if not is_abstract:
            # This is the only time the Statics are ever actually instantiated, and it's only to produce the kwargs
            # passed to the underlying dataclass.
            tmp_inst = cls()
            inst_kw = dc.asdict(tmp_inst)  # type: ignore[call-overload]  # noqa
            inst = sdc_cls(**inst_kw)

            cls.__static_dataclass_instance__ = inst

            # Make all field values available via static `Class.field` access, even those created via a factory. Note
            # that further inheritance of this non-abstract Static will continue to inherit dc.Field instances
            # (including things like their metadata) via dc.fields() access, which does not reference `cls.__dict__`.
            for fld in dc_flds:
                v = getattr(inst, fld.name)
                setattr(cls, fld.name, v)

            def __new__(new_cls, *new_args, **new_kwargs):  # noqa
                try:
                    return new_cls.__dict__['__static_dataclass_instance__']
                except KeyError:
                    return super().__new__(new_cls)

            cls.__new__ = __new__  # type: ignore

        cls.__init__ = Static.__init__  # type: ignore

    @ta.final
    def __init__(self) -> None:
        # This stub serves to allow `StaticSubclass()` to typecheck by allowing it to accept only zero arguments. Note
        # that this is only the case when `Static` is first in mro.
        raise TypeError('May not instantiate static dataclasses')



import configparser

from ..codecs import make_object_lazy_loaded_codec
from ..codecs import make_str_object_codec
from .sections import IniSectionSettingsMap
from .sections import extract_ini_sections
from .sections import render_ini_sections


##


def loads(s: str) -> IniSectionSettingsMap:
    cp = configparser.ConfigParser()
    cp.read_string(s)
    return extract_ini_sections(cp)


def dumps(sm: IniSectionSettingsMap) -> str:
    return render_ini_sections(sm)


INI_CODEC = make_str_object_codec('ini', dumps, loads)

# @omlish-manifest
_INI_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'INI_CODEC', INI_CODEC)



import typing as ta

from .consts import MAX_LEN_CLIENT_TAG_DATA
from .consts import MAX_LEN_TAG_DATA
from .errors import BadCharactersError
from .errors import LineEmptyError
from .errors import TagsTooLongError
from .message import Message
from .tags import parse_tags
from .utils import is_ascii
from .utils import trim_initial_spaces
from .utils import truncate_utf8_safe


class ParsedLine(ta.NamedTuple):
    message: Message

    truncated: bool = False


def parse_line_(
        line: str,
        *,
        max_tag_data_length: int | None = None,
        truncate_len: int | None = None,
) -> ParsedLine:
    # Remove either \n or \r\n from the end of the line:
    line = line.removesuffix('\n')
    line = line.removesuffix('\r')

    # Whether we removed them ourselves, or whether they were removed previously, they count against the line limit:
    if truncate_len is not None:
        if truncate_len <= 2:
            raise LineEmptyError
        truncate_len -= 2

    # Now validate for the 3 forbidden bytes:
    if any(c in line for c in '\x00\n\r'):
        raise BadCharactersError

    if not line:
        raise LineEmptyError

    #

    # Handle tags
    tags: ta.Mapping[str, str] | None = None
    if line.startswith('@'):
        tag_end = line.find(' ')
        if tag_end == -1:
            raise LineEmptyError
        raw_tags = line[1:tag_end]
        if max_tag_data_length is not None and len(raw_tags) > max_tag_data_length:
            raise TagsTooLongError
        tags = parse_tags(raw_tags)
        # Skip over the tags and the separating space
        line = line[tag_end + 1:]

    #

    # Truncate if desired
    truncated = False
    if truncate_len is not None and len(line) > truncate_len:
        line = truncate_utf8_safe(line, truncate_len)
        truncated = True

    line = trim_initial_spaces(line)

    # Handle source
    source: str | None = None

    if line.startswith(':'):
        source_end = line.find(' ')
        if source_end == -1:
            raise LineEmptyError
        source = line[1:source_end]
        line = line[source_end + 1:]

    # Modern: "These message parts, and parameters themselves, are separated by one or more ASCII SPACE characters"
    line = trim_initial_spaces(line)

    # Handle command
    command_end = line.find(' ')
    param_start = command_end + 1 if command_end != -1 else len(line)
    base_command = line[:command_end] if command_end != -1 else line

    if not base_command:
        raise LineEmptyError
    # Technically this must be either letters or a 3-digit numeric:
    if not is_ascii(base_command):
        raise BadCharactersError

    # Normalize command to uppercase:
    command = base_command.upper()
    line = line[param_start:]

    # Handle parameters
    params: list[str] = []
    while line:
        line = trim_initial_spaces(line)
        if not line:
            break
        # Handle trailing
        if line.startswith(':'):
            params.append(line[1:])
            break
        param_end = line.find(' ')
        if param_end == -1:
            params.append(line)
            break
        params.append(line[:param_end])
        line = line[param_end + 1:]

    #

    msg = Message(
        source=source,
        command=command,
        params=params,
        tags=tags,
    )

    return ParsedLine(
        msg,
        truncated=truncated,
    )


def parse_line(
        line: str,
        *,
        max_tag_data_length: int | None = None,
        truncate_len: int | None = None,
) -> Message:
    return parse_line_(
        line,
        max_tag_data_length=max_tag_data_length,
        truncate_len=truncate_len,
    ).message


def parse_line_strict(
        line: str,
        from_client: bool,
        truncate_len: int | None,
) -> ParsedLine:
    max_tag_data_length = MAX_LEN_TAG_DATA
    if from_client:
        max_tag_data_length = MAX_LEN_CLIENT_TAG_DATA

    return parse_line_(
        line,
        max_tag_data_length=max_tag_data_length,
        truncate_len=truncate_len,
    )



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import os
import typing as ta

from ..lite.check import check


@dc.dataclass(frozen=True)
class FileMode:
    """
    https://docs.python.org/3/library/functions.html#open

    'r' open for reading (default)
    'w' open for writing, truncating the file first
    'x' open for exclusive creation, failing if the file already exists
    'a' open for writing, appending to the end of file if it exists
    'b' binary mode
    't' text mode (default)
    '+' open for updating (reading and writing)

    ==

    https://en.cppreference.com/w/cpp/io/c/fopen

    "r"   read             Open a file for reading        read from start   return NULL and set error
    "w"   write            Create a file for writing     destroy contents  create new
    "a"   append           Append to a file              write to end      create new
    "r+"  read extended    Open a file for read/write    read from start   return NULL and set error
    "w+"  write extended   Create a file for read/write  destroy contents  create new
    "a+"  append extended  Open a file for read/write    write to end      create new

    ==

    a (prs): w a
    a (new): w
    a+ (prs): r w a
    a+ (new): r w

    x (new): w
    x+ (new): r w

    w (prs): w
    w (new): w
    w+ (prs): r w
    w+ (new): r w

    r (prs): r
    r+ (prs): r w
    """

    read: bool
    write: bool

    create: bool
    exists: ta.Literal['beginning', 'truncate', 'fail', 'append']

    binary: bool

    #

    def flags(self) -> int:
        return (
            (
                os.O_RDWR if self.read and self.write else
                os.O_WRONLY if self.write else
                os.O_RDONLY if self.read else
                0
            ) |
            (os.O_CREAT if self.create else 0) |
            (
                os.O_APPEND if self.exists == 'append' else
                os.O_EXCL if self.exists == 'fail' else
                os.O_TRUNC if self.exists == 'truncate' else
                0
            )
        )

    @classmethod
    def from_flags(cls, i: int) -> 'FileMode':
        if i & os.O_RDWR:
            i &= ~os.O_RDWR
            read = write = True
        elif i & os.O_WRONLY:
            i &= ~os.O_WRONLY
            write = True
            read = False
        else:
            read = True
            write = False

        create = False
        if i & os.O_CREAT:
            i &= ~os.O_CREAT
            create = True

        exists: str | None = None
        if i & os.O_TRUNC:
            i &= ~os.O_TRUNC
            exists = check.replacing_none(exists, 'truncate')
        if i & os.O_EXCL:
            i &= ~os.O_EXCL
            exists = check.replacing_none(exists, 'fail')
        if i & os.O_APPEND:
            i &= ~os.O_APPEND
            exists = check.replacing_none(exists, 'append')
        if exists is None:
            exists = 'beginning'

        if i:
            raise ValueError(i)

        return FileMode(
            read=read,
            write=write,
            create=create,
            exists=exists,  # type: ignore
            binary=True,
        )

    #

    def render(self) -> str:
        return ''.join([
            (
                'a' if self.exists == 'append' else
                'x' if self.exists == 'fail' else
                'w' if self.exists == 'truncate' else
                'r'
            ),
            '+' if self.read and self.write else '',
            'b' if self.binary else '',
        ])

    @classmethod
    def parse(cls, s: str) -> 'FileMode':
        rwxa: ta.Literal['r', 'w', 'x', 'a', None] = None
        tb: ta.Literal['t', 'b', None] = None
        p: bool | None = None

        for c in s:
            if c in 'rwxa':
                rwxa = check.replacing_none(rwxa, c)  # type: ignore[arg-type]
            elif c in 'tb':
                tb = check.replacing_none(tb, c)  # type: ignore[arg-type]
            elif c == '+':
                p = check.replacing_none(p, True)
            else:
                raise ValueError(c)

        if rwxa is None:
            rwxa = 'r'
        if tb is None:
            tb = 't'
        p = bool(p)

        return FileMode(
            read=rwxa == 'r' or p,
            write=rwxa != 'r' or p,
            create=rwxa != 'r',
            exists=(  # noqa
                'append' if rwxa == 'a' else
                'fail' if rwxa == 'x' else
                'truncate' if rwxa == 'w' else
                'beginning'
            ),
            binary=tb == 'b',
        )



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import typing as ta

from ..lite.cached import cached_nullary
from ..lite.check import check


@dc.dataclass(frozen=True)
class StringMangler:
    escape: str
    escaped: ta.Sequence[str]

    @classmethod
    def of(cls, escape: str, escaped: ta.Iterable[str]) -> 'StringMangler':
        check.arg(len(escape) == 1)
        return StringMangler(escape, sorted(set(escaped) - {escape}))

    def __post_init__(self) -> None:
        check.non_empty_str(self.escape)
        check.arg(len(self.escape) == 1)
        check.not_in(self.escape, self.escaped)
        check.arg(len(set(self.escaped)) == len(self.escaped))

    @cached_nullary
    def replacements(self) -> ta.Sequence[ta.Tuple[str, str]]:
        return [(l, self.escape + str(i)) for i, l in enumerate([self.escape, *self.escaped])]

    def mangle(self, s: str) -> str:
        for l, r in self.replacements():
            s = s.replace(l, r)
        return s

    def unmangle(self, s: str) -> str:
        for l, r in reversed(self.replacements()):
            s = s.replace(r, l)
        return s



"""
TODO:
 - max buf size
 - max recursion depth
 - mark start pos of tokens, currently returning end
 - _do_string inner loop optimization somehow
 - json5 mode
"""
import dataclasses as dc
import io
import json
import re
import typing as ta

from .... import check
from ....funcs.genmachine import GenMachine
from .errors import JsonStreamError


##


ValueTokenKind: ta.TypeAlias = ta.Literal[
    'STRING',
    'NUMBER',

    'SPECIAL_NUMBER',
    'BOOLEAN',
    'NULL',
]

VALUE_TOKEN_KINDS = frozenset(check.isinstance(a, str) for a in ta.get_args(ValueTokenKind))

ControlTokenKind: ta.TypeAlias = ta.Literal[
    'LBRACE',
    'RBRACE',
    'LBRACKET',
    'RBRACKET',
    'COMMA',
    'COLON',
]

TokenKind: ta.TypeAlias = ValueTokenKind | ControlTokenKind

#

ScalarValue: ta.TypeAlias = str | float | int | None

SCALAR_VALUE_TYPES: tuple[type, ...] = tuple(
    check.isinstance(e, type) if e is not None else type(None)
    for e in ta.get_args(ScalarValue)
)


##


class Position(ta.NamedTuple):
    ofs: int
    line: int
    col: int


class Token(ta.NamedTuple):
    kind: TokenKind
    value: ScalarValue
    raw: str | None

    pos: Position

    def __iter__(self):
        raise TypeError


NUMBER_PAT = re.compile(r'-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?')

CONTROL_TOKENS: ta.Mapping[str, TokenKind] = {
    '{': 'LBRACE',
    '}': 'RBRACE',
    '[': 'LBRACKET',
    ']': 'RBRACKET',
    ',': 'COMMA',
    ':': 'COLON',
}

CONST_TOKENS: ta.Mapping[str, tuple[TokenKind, str | float | None]] = {
    'NaN': ('SPECIAL_NUMBER', float('nan')),
    'Infinity': ('SPECIAL_NUMBER', float('inf')),
    '-Infinity': ('SPECIAL_NUMBER', float('-inf')),

    'true': ('BOOLEAN', True),
    'false': ('BOOLEAN', False),
    'null': ('NULL', None),
}


##


@dc.dataclass()
class JsonStreamLexError(JsonStreamError):
    message: str

    pos: Position


class JsonStreamLexer(GenMachine[str, Token]):
    def __init__(
            self,
            *,
            include_raw: bool = False,
    ) -> None:
        self._include_raw = include_raw

        self._ofs = 0
        self._line = 1
        self._col = 0

        self._buf = io.StringIO()

        super().__init__(self._do_main())

    @property
    def pos(self) -> Position:
        return Position(
            self._ofs,
            self._line,
            self._col,
        )

    def _char_in(self, c: str) -> str:
        if c and len(c) != 1:
            raise ValueError(c)

        self._ofs += 1

        if c == '\n':
            self._line += 1
            self._col = 0
        else:
            self._col += 1

        return c

    def _make_tok(
            self,
            kind: TokenKind,
            value: ScalarValue,
            raw: str,
            pos: Position,
    ) -> ta.Sequence[Token]:
        tok = Token(
            kind,
            value,
            raw if self._include_raw else None,
            pos,
        )
        return (tok,)

    def _flip_buf(self) -> str:
        raw = self._buf.getvalue()
        self._buf.seek(0)
        self._buf.truncate()
        return raw

    def _raise(self, msg: str, src: Exception | None = None) -> ta.NoReturn:
        raise JsonStreamLexError(msg, self.pos) from src

    def _do_main(self):
        while True:
            c = self._char_in((yield None))  # noqa

            if not c:
                return None

            if c.isspace():
                continue

            if c in CONTROL_TOKENS:
                yield self._make_tok(CONTROL_TOKENS[c], c, c, self.pos)
                continue

            if c == '"':
                return self._do_string()

            if c.isdigit() or c == '-':
                return self._do_number(c)

            if c in 'tfnIN':
                return self._do_const(c)

            self._raise(f'Unexpected character: {c}')

    def _do_string(self):
        check.state(self._buf.tell() == 0)
        self._buf.write('"')

        pos = self.pos

        last = None
        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c:
                self._raise(f'Unterminated string literal: {self._buf.getvalue()}')

            self._buf.write(c)
            if c == '"' and last != '\\':
                break
            last = c

        raw = self._flip_buf()
        try:
            sv = json.loads(raw)
        except json.JSONDecodeError as e:
            self._raise(f'Invalid string literal: {raw!r}', e)

        yield self._make_tok('STRING', sv, raw, pos)

        return self._do_main()

    def _do_number(self, c: str):
        check.state(self._buf.tell() == 0)
        self._buf.write(c)

        pos = self.pos

        while True:
            try:
                c = self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c:
                break

            if not (c.isdigit() or c in '.eE+-'):
                break
            self._buf.write(c)

        raw = self._flip_buf()

        #

        if not NUMBER_PAT.fullmatch(raw):
            # Can only be -Infinity

            if not c:
                self._raise('Unexpected end of input')

            raw += c
            try:
                for _ in range(7):
                    raw += self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if raw != '-Infinity':
                self._raise(f'Invalid number format: {raw}')

            tk, tv = CONST_TOKENS[raw]
            yield self._make_tok(tk, tv, raw, pos)

            return self._do_main()

        #

        if '.' in raw or 'e' in raw or 'E' in raw:
            nv = float(raw)
        else:
            nv = int(raw)
        yield self._make_tok('NUMBER', nv, raw, pos)

        #

        if not c:
            return None

        if c in CONTROL_TOKENS:
            yield self._make_tok(CONTROL_TOKENS[c], c, c, pos)

        elif not c.isspace():
            self._raise(f'Unexpected character after number: {c}')

        return self._do_main()

    def _do_const(self, c: str):
        pos = self.pos
        raw = c
        while True:
            try:
                raw += self._char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if raw in CONST_TOKENS:
                break

            if len(raw) > 8:  # None of the keywords are longer than 8 characters
                self._raise(f'Invalid literal: {raw}')

        tk, tv = CONST_TOKENS[raw]
        yield self._make_tok(tk, tv, raw, pos)

        return self._do_main()



# ruff: noqa: UP006 UP007
# @omlish-lite
import socket
import threading
import typing as ta

from ..lite.timeouts import Timeout
from ..lite.timeouts import TimeoutLike


##


def socket_can_connect(
        address: ta.Any,
        *,
        timeout: ta.Optional[TimeoutLike] = None,
        on_fail: ta.Optional[ta.Callable[[BaseException], None]] = None,
        exception: ta.Union[ta.Type[BaseException], ta.Tuple[ta.Type[BaseException], ...]] = (ConnectionRefusedError,),
) -> bool:
    timeout = Timeout.of(timeout)

    try:
        conn = socket.create_connection(address, timeout=timeout.or_(None))

    except exception as e:  # noqa
        if on_fail is not None:
            on_fail(e)
        return False

    else:
        conn.close()
        return True


def socket_wait_until_can_connect(
        address: ta.Any,
        *,
        timeout: ta.Optional[TimeoutLike] = None,
        on_fail: ta.Optional[ta.Callable[[BaseException], None]] = None,
        sleep_s: float = .1,
        exception: ta.Union[ta.Type[BaseException], ta.Tuple[ta.Type[BaseException], ...]] = (ConnectionRefusedError,),
        cancel_event: ta.Optional[threading.Event] = None,
) -> None:
    timeout = Timeout.of(timeout)

    if cancel_event is None:
        cancel_event = threading.Event()

    while not cancel_event.is_set():
        timeout()

        if socket_can_connect(
            address,
            timeout=timeout,
            on_fail=on_fail,
            exception=exception,
        ):
            break

        cancel_event.wait(min(sleep_s, timeout.remaining()))



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import os
import struct
import sys
import typing as ta


@dc.dataclass(frozen=True)
class FcntlLockData:
    # cmd = {F_SETLK, F_SETLKW, F_GETLK}

    type: int  # {F_RDLCK, F_WRLCK, F_UNLCK}
    whence: int = os.SEEK_SET
    start: int = 0
    len: int = 0
    pid: int = 0

    #

    _STRUCT_PACKING_BY_PLATFORM: ta.ClassVar[ta.Mapping[str, ta.Sequence[ta.Tuple[str, str]]]] = {
        'linux': [
            ('type', 'h'),
            ('whence', 'h'),
            ('start', 'q'),
            ('len', 'q'),
            ('pid', 'i'),
        ],
        'darwin': [
            ('start', 'q'),
            ('len', 'q'),
            ('pid', 'i'),
            ('type', 'h'),
            ('whence', 'h'),
        ],
    }

    @classmethod
    def _struct_packing(cls) -> ta.Sequence[ta.Tuple[str, str]]:
        try:
            return cls._STRUCT_PACKING_BY_PLATFORM[sys.platform]
        except KeyError:
            raise OSError from None

    def pack(self) -> bytes:
        packing = self._struct_packing()
        fmt = ''.join(f for _, f in packing)
        tup = [getattr(self, a) for a, _ in packing]
        return struct.pack(fmt, *tup)

    @classmethod
    def unpack(cls, data: bytes) -> 'FcntlLockData':
        packing = cls._struct_packing()
        fmt = ''.join(f for _, f in packing)
        tup = struct.unpack(fmt, data)
        kw = {a: v for (a, _), v in zip(packing, tup)}
        return FcntlLockData(**kw)



import dataclasses as dc
import typing as ta

from ... import lang
from .internals import FIELDS_ATTR
from .internals import PARAMS_ATTR
from .processing import Processor
from .reflect import ClassInfo
from .utils import Namespace
from .utils import create_fn
from .utils import set_new_attribute


if ta.TYPE_CHECKING:
    from . import metaclass
else:
    metaclass = lang.proxy_import('.metaclass', __package__)


def check_frozen_bases(info: ClassInfo) -> None:
    mc_base = getattr(metaclass, 'Data', None)
    all_frozen_bases = None
    any_frozen_base = False
    has_dataclass_bases = False
    for b in info.cls.__mro__[-1:0:-1]:
        if b is mc_base:
            continue
        base_fields = getattr(b, FIELDS_ATTR, None)
        if base_fields is not None:
            has_dataclass_bases = True
            if all_frozen_bases is None:
                all_frozen_bases = True
            current_frozen = getattr(b, PARAMS_ATTR).frozen
            all_frozen_bases = all_frozen_bases and current_frozen
            any_frozen_base = any_frozen_base or current_frozen

    if has_dataclass_bases:
        if any_frozen_base and not info.params.frozen:
            raise TypeError('cannot inherit non-frozen dataclass from a frozen one')

        if all_frozen_bases is False and info.params.frozen:
            raise TypeError('cannot inherit frozen dataclass from a non-frozen one')


def frozen_get_del_attr(
        cls: type,
        fields: ta.Sequence[dc.Field],
        globals: Namespace,  # noqa
) -> tuple[ta.Callable, ta.Callable]:
    locals = {  # noqa
        'cls': cls,
        'FrozenInstanceError': dc.FrozenInstanceError,
    }
    condition = 'type(self) is cls'
    if fields:
        condition += ' or name in {' + ', '.join(repr(f.name) for f in fields) + '}'
    return (
        create_fn(
            '__setattr__',
            ('self', 'name', 'value'),
            [
                f'if {condition}:',
                ' raise FrozenInstanceError(f"cannot assign to field {name!r}")',
                f'super(cls, self).__setattr__(name, value)',
            ],
            locals=locals,
            globals=globals,
        ),
        create_fn(
            '__delattr__',
            ('self', 'name'),
            [
                f'if {condition}:',
                ' raise FrozenInstanceError(f"cannot delete field {name!r}")',
                f'super(cls, self).__delattr__(name)',
            ],
            locals=locals,
            globals=globals,
        ),
    )


class FrozenProcessor(Processor):
    def check(self) -> None:
        check_frozen_bases(self._info)

    def _process(self) -> None:
        if not self._info.params.frozen:
            return

        for fn in frozen_get_del_attr(self._cls, self._info.instance_fields, self._info.globals):
            if set_new_attribute(self._cls, fn.__name__, fn):
                raise TypeError(f'Cannot overwrite attribute {fn.__name__} in class {self._cls.__name__}')



import inspect

from .init import get_init_fields
from .processing import Processor
from .utils import create_fn
from .utils import set_new_attribute


class EqProcessor(Processor):
    def _process(self) -> None:
        if not self._info.params.eq:
            return

        # flds = [f for f in self._info.instance_fields if f.compare]
        # self_tuple = tuple_str('self', flds)
        # other_tuple = tuple_str('other', flds)
        # set_new_attribute(cls, '__eq__', _cmp_fn('__eq__', '==', self_tuple, other_tuple, globals=globals))
        cmp_fields = (field for field in self._info.instance_fields if field.compare)
        terms = [f'self.{field.name} == other.{field.name}' for field in cmp_fields]
        field_comparisons = ' and '.join(terms) or 'True'
        body = [
            f'if self is other:',
            f' return True',
            f'if other.__class__ is self.__class__:',
            f' return {field_comparisons}',
            f'return NotImplemented',
        ]
        func = create_fn('__eq__', ('self', 'other'), body, globals=self._info.globals)
        set_new_attribute(self._cls, '__eq__', func)


class DocProcessor(Processor):
    def _process(self) -> None:
        if getattr(self._cls, '__doc__'):
            return

        try:
            text_sig = str(inspect.signature(self._cls)).replace(' -> None', '')
        except (TypeError, ValueError):
            text_sig = ''
        self._cls.__doc__ = (self._cls.__name__ + text_sig)


class MatchArgsProcessor(Processor):
    def _process(self) -> None:
        if not self._info.params.match_args:
            return

        ifs = get_init_fields(self._info.fields.values())
        set_new_attribute(self._cls, '__match_args__', tuple(f.name for f in ifs.std))



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
Should be kept somewhat lightweight - used in cli entrypoints.

TODO:
 - persisted caching support - {pkg_name: manifests}
"""
import dataclasses as dc
import importlib.machinery
import importlib.resources
import json
import os.path
import threading
import typing as ta

from .types import Manifest


##


class ManifestLoader:
    def __init__(
            self,
            *,
            module_remap: ta.Optional[ta.Mapping[str, str]] = None,
    ) -> None:
        super().__init__()

        self._lock = threading.RLock()

        self._module_remap = module_remap or {}
        self._module_reverse_remap = {v: k for k, v in self._module_remap.items()}

        self._cls_cache: ta.Dict[str, type] = {}
        self._raw_cache: ta.Dict[str, ta.Optional[ta.Sequence[Manifest]]] = {}

    #

    @classmethod
    def from_entry_point(
            cls,
            globals: ta.Mapping[str, ta.Any],  # noqa
            *,
            module_remap: ta.Optional[ta.Mapping[str, str]] = None,
            **kwargs: ta.Any,
    ) -> 'ManifestLoader':
        rm: ta.Dict[str, str] = {}

        if module_remap:
            rm.update(module_remap)

        if '__name__' in globals and '__spec__' in globals:
            name: str = globals['__name__']
            spec: importlib.machinery.ModuleSpec = globals['__spec__']
            if '__main__' not in rm and name == '__main__':
                rm[spec.name] = '__main__'

        return cls(module_remap=rm, **kwargs)

    #

    def _load_cls(self, key: str) -> type:
        try:
            return self._cls_cache[key]
        except KeyError:
            pass

        if not key.startswith('$'):
            raise Exception(f'Bad key: {key}')

        parts = key[1:].split('.')
        pos = next(i for i, p in enumerate(parts) if p[0].isupper())

        mod_name = '.'.join(parts[:pos])
        mod_name = self._module_remap.get(mod_name, mod_name)
        mod = importlib.import_module(mod_name)

        obj: ta.Any = mod
        for ca in parts[pos:]:
            obj = getattr(obj, ca)

        cls = obj
        if not isinstance(cls, type):
            raise TypeError(cls)

        self._cls_cache[key] = cls
        return cls

    def load_cls(self, key: str) -> type:
        with self._lock:
            return self._load_cls(key)

    #

    def _load_contents(self, obj: ta.Any, pkg_name: str) -> ta.Sequence[Manifest]:
        if not isinstance(obj, (list, tuple)):
            raise TypeError(obj)

        lst: ta.List[Manifest] = []
        for e in obj:
            m = Manifest(**e)

            m = dc.replace(m, module=pkg_name + m.module)

            [(key, value_dct)] = m.value.items()
            if not key.startswith('$'):
                raise Exception(f'Bad key: {key}')
            if key.startswith('$.'):
                key = f'${pkg_name}{key[1:]}'
                m = dc.replace(m, value={key: value_dct})

            lst.append(m)

        return lst

    def load_contents(self, obj: ta.Any, pkg_name: str) -> ta.Sequence[Manifest]:
        with self._lock:
            return self.load_contents(obj, pkg_name)

    #

    def _load_raw(self, pkg_name: str) -> ta.Optional[ta.Sequence[Manifest]]:
        try:
            return self._raw_cache[pkg_name]
        except KeyError:
            pass

        t = importlib.resources.files(pkg_name).joinpath('.manifests.json')
        if not t.is_file():
            self._raw_cache[pkg_name] = None
            return None

        src = t.read_text('utf-8')
        obj = json.loads(src)
        if not isinstance(obj, (list, tuple)):
            raise TypeError(obj)

        lst = self._load_contents(obj, pkg_name)

        self._raw_cache[pkg_name] = lst
        return lst

    def load_raw(self, pkg_name: str) -> ta.Optional[ta.Sequence[Manifest]]:
        with self._lock:
            return self._load_raw(pkg_name)

    #

    def _load(
            self,
            *pkg_names: str,
            only: ta.Optional[ta.Iterable[type]] = None,
    ) -> ta.Sequence[Manifest]:
        only_keys: ta.Optional[ta.Set]
        if only is not None:
            only_keys = set()
            for cls in only:
                if not (isinstance(cls, type) and dc.is_dataclass(cls)):
                    raise TypeError(cls)
                mod_name = cls.__module__
                mod_name = self._module_reverse_remap.get(mod_name, mod_name)
                only_keys.add(f'${mod_name}.{cls.__qualname__}')
        else:
            only_keys = None

        lst: ta.List[Manifest] = []
        for pn in pkg_names:
            for manifest in (self.load_raw(pn) or []):
                [(key, value_dct)] = manifest.value.items()
                if only_keys is not None and key not in only_keys:
                    continue

                cls = self._load_cls(key)
                value = cls(**value_dct)

                manifest = dc.replace(manifest, value=value)
                lst.append(manifest)

        return lst

    def load(
            self,
            *pkg_names: str,
            only: ta.Optional[ta.Iterable[type]] = None,
    ) -> ta.Sequence[Manifest]:
        with self._lock:
            return self._load(
                *pkg_names,
                only=only,
            )

    #

    ENTRY_POINT_GROUP = 'omlish.manifests'

    def discover_pkgs(self) -> ta.Sequence[str]:
        # This is a fat dep so do it late.
        importlib_metadata = __import__('importlib.metadata').metadata

        return [
            ep.value
            for ep in importlib_metadata.entry_points(group=self.ENTRY_POINT_GROUP)
        ]

    def scan_pkg_root(self, root: str) -> ta.Sequence[str]:
        pkgs: ta.List[str] = []
        for n in os.listdir(root):
            if os.path.isdir(p := os.path.join(root, n)) and os.path.exists(os.path.join(p, '__init__.py')):
                pkgs.append(n)
        return pkgs

    def scan_or_discover_pkgs(
            self,
            *,
            specified_roots: ta.Optional[ta.Sequence[str]] = None,
            fallback_root: ta.Optional[str] = None,
    ) -> ta.Sequence[str]:
        pkgs: list[str] = []

        if specified_roots is not None:
            if isinstance(specified_roots, str):
                raise TypeError(specified_roots)
            for r in specified_roots:
                pkgs.extend(self.scan_pkg_root(r))

        else:
            pkgs.extend(self.discover_pkgs())

            if not pkgs and fallback_root is not None:
                pkgs.extend(self.scan_pkg_root(fallback_root))

        return pkgs


##


MANIFEST_LOADER = ManifestLoader()



from .base import (  # noqa
    Deathpact,
    NopDeathpact,

    BaseDeathpact,
)

from .heartbeatfile import (  # noqa
    HeartbeatFileDeathpact,
)

from .pipe import (  # noqa
    PipeDeathpact,
    ForkAwarePipeDeathpact,
)



import inspect
import typing as ta

from ... import check
from ... import collections as col
from ... import lang
from ... import reflect as rfl
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactory
from ..base import Option
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from .marshal import ObjectMarshaler
from .metadata import FieldInfo
from .metadata import FieldInfos
from .metadata import FieldMetadata
from .unmarshal import ObjectUnmarshaler


##


def _is_namedtuple(rty: rfl.Type) -> bool:
    return (
        isinstance(rty, type) and
        issubclass(rty, tuple) and
        ta.NamedTuple in rfl.get_orig_bases(rty)
    )


def get_namedtuple_field_infos(
        ty: type,
        opts: col.TypeMap[Option] = col.TypeMap(),
) -> FieldInfos:
    check.arg(_is_namedtuple(ty), ty)

    sig = inspect.signature(ty)

    ret: list[FieldInfo] = []
    for param in sig.parameters.values():
        ret.append(FieldInfo(
            name=param.name,
            type=param.annotation,
            metadata=FieldMetadata(),

            marshal_name=param.name,
            unmarshal_names=[param.name],
        ))

    return FieldInfos(ret)


##


class NamedtupleMarshalerFactory(MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return _is_namedtuple(rty)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        check.state(_is_namedtuple(rty))
        ty = check.isinstance(rty, type)
        check.state(not lang.is_abstract_class(ty))

        fis = get_namedtuple_field_infos(ty, ctx.options)

        fields = [
            (fi, ctx.make(fi.type))
            for fi in fis
        ]

        return ObjectMarshaler(
            fields,
        )


##


class NamedtupleUnmarshalerFactory(UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return _is_namedtuple(rty)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        check.state(_is_namedtuple(rty))
        ty = check.isinstance(rty, type)
        check.state(not lang.is_abstract_class(ty))

        fis = get_namedtuple_field_infos(ty, ctx.options)

        d: dict[str, tuple[FieldInfo, Unmarshaler]] = {}
        defaults: dict[str, ta.Any] = {}

        for fi in fis:
            tup = (fi, ctx.make(fi.type))

            for un in fi.unmarshal_names:
                if un in d:
                    raise KeyError(f'Duplicate fields for name {un!r}: {fi.name!r}, {d[un][0].name!r}')
                d[un] = tup

            if fi.options.default.present:
                defaults[fi.name] = fi.options.default.must()

        return ObjectUnmarshaler(
            ty,
            d,
            defaults=defaults,
        )



import dataclasses as dc
import enum
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from ..values import Value


@dc.dataclass(frozen=True)
class EnumMarshaler(Marshaler):
    ty: type[enum.Enum]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return o.name


class EnumMarshalerFactory(MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and issubclass(rty, enum.Enum)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        ty = check.isinstance(rty, type)
        check.state(issubclass(ty, enum.Enum))
        return EnumMarshaler(ty)  # noqa


@dc.dataclass(frozen=True)
class EnumUnmarshaler(Unmarshaler):
    ty: type[enum.Enum]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self.ty[check.isinstance(v, str)]


class EnumUnmarshalerFactory(UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and issubclass(rty, enum.Enum)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        ty = check.isinstance(rty, type)
        check.state(issubclass(ty, enum.Enum))
        return EnumUnmarshaler(ty)



import dataclasses as dc
import typing as ta

from ... import reflect as rfl
from ...funcs import match as mfs
from ..base import MarshalContext
from ..base import Marshaler
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..exceptions import ForbiddenTypeError


C = ta.TypeVar('C')
R = ta.TypeVar('R')


@dc.dataclass(frozen=True)
class ForbiddenTypeFactory(mfs.MatchFn[[C, rfl.Type], R]):
    rtys: ta.AbstractSet[rfl.Type]

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        return rty in self.rtys

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        raise ForbiddenTypeError(rty)


@dc.dataclass(frozen=True)
class ForbiddenTypeMarshalerFactory(ForbiddenTypeFactory[MarshalContext, Marshaler]):
    pass


@dc.dataclass(frozen=True)
class ForbiddenTypeUnmarshalerFactory(ForbiddenTypeFactory[UnmarshalContext, Unmarshaler]):
    pass



import typing as ta

from ... import reflect as rfl
from ..base import MarshalContext
from ..base import Marshaler
from ..base import TypeMapMarshalerFactory
from ..base import TypeMapUnmarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value


class AnyMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return ctx.make(type(o)).marshal(ctx, o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return v


ANY_MARSHALER_UNMARSHALER = AnyMarshalerUnmarshaler()

ANY_MARSHALER_FACTORY = TypeMapMarshalerFactory({rfl.ANY: ANY_MARSHALER_UNMARSHALER})
ANY_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({rfl.ANY: ANY_MARSHALER_UNMARSHALER})



"""
TODO:
 - async<->sync greeenlet bridge
  In [5]: %timeit greenlet.greenlet(f).switch()
  517 ns ± 13.2 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
  - injected io provider - sync vs greenlet aio trampolined
 - push/pull bridge?
 - move to lang

https://github.com/sqlalchemy/sqlalchemy/blob/1e75c189da721395bc8c2d899c722a5b9a170404/lib/sqlalchemy/util/_concurrency_py3k.py#L83
"""
import functools
import typing as ta


T = ta.TypeVar('T')


def sync_await(fn: ta.Callable[..., T], *args: ta.Any, **kwargs: ta.Any) -> T:
    ret = missing = object()

    async def gate():
        nonlocal ret
        ret = await fn(*args, **kwargs)  # type: ignore

    cr = gate()
    try:
        try:
            cr.send(None)
        except StopIteration:
            pass
        if ret is missing or cr.cr_await is not None or cr.cr_running:
            raise TypeError('Not terminated')
    finally:
        cr.close()

    return ret  # type: ignore


def sync_list(fn: ta.Callable[..., ta.AsyncIterator[T]], *args, **kwargs) -> list[T]:
    lst = None

    async def inner():
        nonlocal lst
        lst = [v async for v in fn(*args, **kwargs)]

    sync_await(inner)
    if not isinstance(lst, list):
        raise TypeError(lst)
    return lst


async def async_list(fn: ta.Callable[..., ta.AsyncIterator[T]], *args, **kwargs) -> list[T]:
    return [v async for v in fn(*args, **kwargs)]


class SyncableIterable(ta.Generic[T]):
    def __init__(self, obj) -> None:
        super().__init__()
        self._obj = obj

    def __iter__(self) -> ta.Iterator[T]:
        async def inner():
            async for i in self._obj:
                yield i
        return iter(sync_list(inner))

    def __aiter__(self) -> ta.AsyncIterator[T]:
        return self._obj.__aiter__()


def syncable_iterable(fn: ta.Callable[..., ta.AsyncIterator[T]]) -> ta.Callable[..., SyncableIterable[T]]:
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        return SyncableIterable(fn(*args, **kwargs))
    return inner



import abc
import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactory
from ..values import Value
from .metadata import FieldTypeTagging
from .metadata import Impls
from .metadata import Polymorphism
from .metadata import TypeTagging
from .metadata import WrapperTypeTagging


##


class PolymorphismMarshaler(Marshaler, lang.Abstract):
    @abc.abstractmethod
    def get_impls(self) -> ta.Mapping[type, tuple[str, Marshaler]]:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class WrapperPolymorphismMarshaler(PolymorphismMarshaler):
    m: ta.Mapping[type, tuple[str, Marshaler]]

    def get_impls(self) -> ta.Mapping[type, tuple[str, Marshaler]]:
        return self.m

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        tag, m = self.m[type(o)]
        return {tag: m.marshal(ctx, o)}


@dc.dataclass(frozen=True)
class FieldPolymorphismMarshaler(PolymorphismMarshaler):
    m: ta.Mapping[type, tuple[str, Marshaler]]
    tf: str

    def get_impls(self) -> ta.Mapping[type, tuple[str, Marshaler]]:
        return self.m

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        tag, m = self.m[type(o)]
        return {self.tf: tag, **m.marshal(ctx, o)}  # type: ignore


def make_polymorphism_marshaler(
        impls: Impls,
        tt: TypeTagging,
        ctx: MarshalContext,
) -> Marshaler:
    m = {
        i.ty: (i.tag, ctx.make(i.ty))
        for i in impls
    }
    if isinstance(tt, WrapperTypeTagging):
        return WrapperPolymorphismMarshaler(m)
    elif isinstance(tt, FieldTypeTagging):
        return FieldPolymorphismMarshaler(m, tt.field)
    else:
        raise TypeError(tt)


@dc.dataclass(frozen=True)
class PolymorphismMarshalerFactory(MarshalerFactory):
    p: Polymorphism
    tt: TypeTagging = WrapperTypeTagging()

    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return rty is self.p.ty

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        check.is_(rty, self.p.ty)
        return make_polymorphism_marshaler(self.p.impls, self.tt, ctx)



import dataclasses as dc
import reprlib
import typing as ta

from .params import get_field_extras
from .processing import Processor
from .utils import Namespace
from .utils import create_fn
from .utils import set_new_attribute


def repr_fn(
        fields: ta.Sequence[dc.Field],
        globals: Namespace,  # noqa
) -> ta.Callable:
    locals: dict[str, ta.Any] = {}  # noqa
    if any(get_field_extras(f).repr_fn is not None for f in fields):
        lst: list[str] = []
        for f in fields:
            if (fex := get_field_extras(f)).repr_fn is not None:
                locals[fn_name := f'__repr_fn__{f.name}'] = fex.repr_fn
                lst.append(f"if (r := {fn_name}(self.{f.name})) is not None: l.append(f'{f.name}={{r}}')")
            else:
                lst.append(f"l.append(f'{f.name}={{self.{f.name}!r}}')")
        src = [
            'l = []',
            *lst,
            'return f"{self.__class__.__qualname__}({", ".join(l)})"',
        ]
    else:
        src = [
            'return f"{self.__class__.__qualname__}(' +
            ', '.join([f'{f.name}={{self.{f.name}!r}}' for f in fields]) +
            ')"',
        ]
    fn = create_fn(
        '__repr__',
        ('self',),
        src,
        globals=globals,
        locals=locals,
    )
    return reprlib.recursive_repr()(fn)


class ReprProcessor(Processor):
    def _process(self) -> None:
        if not self._info.params.repr:
            return

        flds = [f for f in self._info.instance_fields if f.repr]
        set_new_attribute(self._cls, '__repr__', repr_fn(flds, self._info.globals))  # noqa



"""
TODO:
 - chaining
"""
import os
import time
import typing as ta

from ... import check
from ..forkhooks import ProcessOriginTracker
from ..temp import make_temp_file
from .base import BaseDeathpact


##


class HeartbeatFileDeathpact(BaseDeathpact):
    def __init__(
            self,
            path: str,
            ttl_s: float = 10.,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self._path = path
        self._ttl_s = ttl_s

        self._process_origin = ProcessOriginTracker()

    @property
    def path(self) -> str:
        return self._path

    def is_parent(self) -> bool:
        return self._process_origin.is_in_origin_process()

    #

    def __enter__(self) -> ta.Self:
        check.state(self.is_parent())

        self.update()

        return self

    def close(self) -> None:
        if self.is_parent():
            try:
                os.unlink(self._path)
            except FileNotFoundError:
                pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    #

    @classmethod
    def _now(cls) -> float:
        return time.monotonic()

    def update(self) -> None:
        check.state(self.is_parent())

        new = make_temp_file()
        with open(new, 'w') as f:
            f.write(str(self._now()))

        # FIXME: same filesystem
        os.replace(new, self._path)

    def read(self) -> float:
        try:
            with open(self._path) as f:
                return float(f.read())
        except FileNotFoundError:
            return float('-inf')

    def age(self) -> float:
        return self._now() - self.read()

    def should_die(self) -> bool:
        return self.age() >= self._ttl_s



# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import dataclasses as dc
import typing as ta

from ..lite.check import check
from ..lite.timeouts import TimeoutLike


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class SubprocessRunOutput(ta.Generic[T]):
    proc: T

    returncode: int  # noqa

    stdout: ta.Optional[bytes] = None
    stderr: ta.Optional[bytes] = None


##


@dc.dataclass(frozen=True)
class SubprocessRun:
    cmd: ta.Sequence[str]
    input: ta.Any = None
    timeout: ta.Optional[TimeoutLike] = None
    check: bool = False
    capture_output: ta.Optional[bool] = None
    kwargs: ta.Optional[ta.Mapping[str, ta.Any]] = None

    #

    _FIELD_NAMES: ta.ClassVar[ta.FrozenSet[str]]

    def replace(self, **kwargs: ta.Any) -> 'SubprocessRun':
        if not kwargs:
            return self

        field_kws = {}
        extra_kws = {}
        for k, v in kwargs.items():
            if k in self._FIELD_NAMES:
                field_kws[k] = v
            else:
                extra_kws[k] = v

        return dc.replace(self, **{
            **dict(kwargs={
                **(self.kwargs or {}),
                **extra_kws,
            }),
            **field_kws,  # passing a kwarg named 'kwargs' intentionally clobbers
        })

    #

    @classmethod
    def of(
            cls,
            *cmd: str,
            input: ta.Any = None,  # noqa
            timeout: ta.Optional[TimeoutLike] = None,
            check: bool = False,  # noqa
            capture_output: ta.Optional[bool] = None,
            **kwargs: ta.Any,
    ) -> 'SubprocessRun':
        return cls(
            cmd=cmd,
            input=input,
            timeout=timeout,
            check=check,
            capture_output=capture_output,
            kwargs=kwargs,
        )

    #

    _DEFAULT_SUBPROCESSES: ta.ClassVar[ta.Optional[ta.Any]] = None  # AbstractSubprocesses

    def run(
            self,
            subprocesses: ta.Optional[ta.Any] = None,  # AbstractSubprocesses
            **kwargs: ta.Any,
    ) -> SubprocessRunOutput:
        if subprocesses is None:
            subprocesses = self._DEFAULT_SUBPROCESSES
        return check.not_none(subprocesses).run_(self.replace(**kwargs))  # type: ignore[attr-defined]

    _DEFAULT_ASYNC_SUBPROCESSES: ta.ClassVar[ta.Optional[ta.Any]] = None  # AbstractAsyncSubprocesses

    async def async_run(
            self,
            async_subprocesses: ta.Optional[ta.Any] = None,  # AbstractAsyncSubprocesses
            **kwargs: ta.Any,
    ) -> SubprocessRunOutput:
        if async_subprocesses is None:
            async_subprocesses = self._DEFAULT_ASYNC_SUBPROCESSES
        return await check.not_none(async_subprocesses).run_(self.replace(**kwargs))  # type: ignore[attr-defined]


SubprocessRun._FIELD_NAMES = frozenset(fld.name for fld in dc.fields(SubprocessRun))  # noqa


##


class SubprocessRunnable(abc.ABC, ta.Generic[T]):
    @abc.abstractmethod
    def make_run(self) -> SubprocessRun:
        raise NotImplementedError

    @abc.abstractmethod
    def handle_run_output(self, output: SubprocessRunOutput) -> T:
        raise NotImplementedError

    #

    def run(
            self,
            subprocesses: ta.Optional[ta.Any] = None,  # AbstractSubprocesses
            **kwargs: ta.Any,
    ) -> T:
        return self.handle_run_output(self.make_run().run(subprocesses, **kwargs))

    async def async_run(
            self,
            async_subprocesses: ta.Optional[ta.Any] = None,  # AbstractAsyncSubprocesses
            **kwargs: ta.Any,
    ) -> T:
        return self.handle_run_output(await self.make_run().async_run(async_subprocesses, **kwargs))



# ruff: noqa: ANN204

class Hashable:
    def __hash__(self): ...


class Iterable:
    def __iter__(self): ...


class Iterator(Iterable):
    def __next__(self): ...


class Reversible(Iterable):
    def __reversed__(self): ...


class Sized:
    def __len__(self): ...


class Container:
    def __contains__(self, x): ...


class Collection(Sized, Iterable, Container):
    pass


# region Sequence


class Sequence(Reversible, Collection):
    def __getitem__(self, index): ...

    def index(self, value, start=0, stop=None): ...

    def count(self, value): ...


class MutableSequence(Sequence):
    def __setitem__(self, index, value): ...

    def __delitem__(self, index): ...

    def insert(self, index, value): ...

    def append(self, value): ...

    def clear(self): ...

    def reverse(self): ...

    def extend(self, values): ...

    def pop(self, index=-1): ...

    def remove(self, value): ...

    def __iadd__(self, values): ...


# endregion


# region Set


class Set(Collection):
    def __le__(self, other): ...

    def __lt__(self, other): ...

    def __gt__(self, other): ...

    def __ge__(self, other): ...

    def __eq__(self, other): ...

    def __and__(self, other): ...

    __rand__ = __and__

    def isdisjoint(self, other): ...

    def __or__(self, other): ...

    __ror__ = __or__

    def __sub__(self, other): ...

    def __rsub__(self, other): ...

    def __xor__(self, other): ...

    __rxor__ = __xor__


class MutableSet(Set):
    def add(self, value): ...

    def discard(self, value): ...

    def remove(self, value): ...

    def pop(self): ...

    def clear(self): ...

    def __ior__(self, it): ...

    def __iand__(self, it): ...

    def __ixor__(self, it): ...

    def __isub__(self, it): ...


# endregion


# region Mapping


class Mapping(Collection):
    def __getitem__(self, key): ...

    def get(self, key, default=None): ...

    def keys(self): ...

    def items(self): ...

    def values(self): ...

    def __eq__(self, other): ...

    __reversed__ = None


class MutableMapping(Mapping):
    def __setitem__(self, key, value): ...

    def __delitem__(self, key): ...

    def pop(self, key, default=None): ...

    def popitem(self): ...

    def clear(self): ...

    def update(self, other=(), **kwds): ...

    def setdefault(self, key, default=None): ...


# endregion



"""
TODO:
 - more logging options
 - a more powerful interface would be run_fn_with_bootstrap..
 - ./python -m gprof2dot -f pstats prof.pstats | dot -Tpdf -o prof.pstats.pdf && open prof.pstats.pdf
 - multiprocess profiling - afterfork, suffix with pid

TODO diag:
 - tracemalloc
 - yappi
 - stackscope
 - https://github.com/pythonspeed/filprofiler
 - https://pypi.org/project/guppy3/
 - https://pypi.org/project/memory-profiler/
 - https://pypi.org/project/Pympler/


TODO new items:
 - packaging fixups
 - daemonize ( https://github.com/thesharp/daemonize/blob/master/daemonize.py )
"""
# ruff: noqa: UP006
import contextlib
import typing as ta

from .. import lang
from . import diag  # noqa
from . import sys  # noqa
from .base import Bootstrap
from .base import ContextBootstrap
from .base import SimpleBootstrap


##


BOOTSTRAP_TYPES_BY_NAME: ta.Mapping[str, ta.Type[Bootstrap]] = {  # noqa
    lang.snake_case(cls.__name__[:-len('Bootstrap')]): cls
    for cls in lang.deep_subclasses(Bootstrap, concrete_only=True)
}

BOOTSTRAP_TYPES_BY_CONFIG_TYPE: ta.Mapping[ta.Type[Bootstrap.Config], ta.Type[Bootstrap]] = {
    cls.Config: cls
    for cls in BOOTSTRAP_TYPES_BY_NAME.values()
}


##


class BootstrapHarness:
    def __init__(self, lst: ta.Sequence[Bootstrap]) -> None:
        super().__init__()
        self._lst = lst

    @contextlib.contextmanager
    def __call__(self) -> ta.Iterator[None]:
        with contextlib.ExitStack() as es:
            for c in self._lst:
                if isinstance(c, SimpleBootstrap):
                    c.run()
                elif isinstance(c, ContextBootstrap):
                    es.enter_context(c.enter())
                else:
                    raise TypeError(c)

            yield


##


@contextlib.contextmanager
def bootstrap(*cfgs: Bootstrap.Config) -> ta.Iterator[None]:
    with BootstrapHarness([
        BOOTSTRAP_TYPES_BY_CONFIG_TYPE[type(c)](c)
        for c in cfgs
    ])():
        yield



import functools
import logging


_log = logging.getLogger(__name__)


def error_logging(log=_log):  # noqa
    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except Exception:
                log.exception('Error in %r', fn)
                raise

        return inner

    return outer



"""
TODO:
 - tangled with objects - Field/ObjectMetadata defined over there but unused
"""
import typing as ta

from ... import check
from ... import collections as col
from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ...lite import marshal as lm
from ..base import MarshalContext
from ..base import Marshaler
from ..base import MarshalerFactory
from ..base import Option
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from ..naming import Naming
from ..naming import translate_name
from .marshal import ObjectMarshaler
from .metadata import DEFAULT_FIELD_OPTIONS
from .metadata import FIELD_OPTIONS_KWARGS
from .metadata import FieldInfo
from .metadata import FieldInfos
from .metadata import FieldMetadata
from .metadata import FieldOptions
from .metadata import ObjectMetadata
from .unmarshal import ObjectUnmarshaler


##


def get_dataclass_metadata(ty: type) -> ObjectMetadata:
    return check.opt_single(
        e
        for e in dc.get_merged_metadata(ty).get(dc.UserMetadata, [])
        if isinstance(e, ObjectMetadata)
    ) or ObjectMetadata()


def get_dataclass_field_infos(
        ty: type,
        opts: col.TypeMap[Option] | None = None,
) -> FieldInfos:
    if opts is None:
        opts = col.TypeMap()

    dc_md = get_dataclass_metadata(ty)
    dc_naming = dc_md.field_naming or opts.get(Naming)
    dc_rf = dc.reflect(ty)

    fi_defaults = {
        k: v
        for k, v in dc.asdict(dc_md.field_defaults).items()
        if v is not None
    }
    fo_defaults = {
        k: v
        for k, v in fi_defaults.pop('options').items()
        if v != getattr(DEFAULT_FIELD_OPTIONS, k)
    }

    type_hints = ta.get_type_hints(ty)

    ret: list[FieldInfo] = []
    for field in dc_rf.fields.values():
        if (f_naming := field.metadata.get(Naming, dc_naming)) is not None:
            um_name = translate_name(field.name, f_naming)
        else:
            um_name = field.name

        f_ty: ta.Any
        if (cpx := dc_rf.cls_params_extras) is not None and cpx.generic_init:
            f_ty = dc_rf.generic_replaced_field_annotations[field.name]
        else:
            f_ty = type_hints[field.name]

        fi_kw = dict(fi_defaults)
        fo_kw = dict(fo_defaults)

        fi_kw.update(
            name=field.name,
            type=f_ty,
            metadata=FieldMetadata(),

            marshal_name=um_name,
            unmarshal_names=[um_name],
        )

        has_set_name = False
        if (fmd := field.metadata.get(FieldMetadata)) is not None:
            fi_kw.update(
                metadata=fmd,
            )

            for fo_k in FIELD_OPTIONS_KWARGS:
                if (fo_v := getattr(fmd.options, fo_k)) != getattr(DEFAULT_FIELD_OPTIONS, fo_k):
                    fo_kw[fo_k] = fo_v

            if fmd.name is not None:
                has_set_name = True
                fi_kw.update(
                    marshal_name=fmd.name,
                    unmarshal_names=col.unique([fmd.name, *(fmd.alts or ())]),
                )

        else:
            try:
                lfk = field.metadata[lm.OBJ_MARSHALER_FIELD_KEY]
            except KeyError:
                pass
            else:
                if lfk is not None:
                    check.non_empty_str(lfk)
                    has_set_name = True
                    fi_kw.update(
                        marshal_name=lfk,
                        unmarshal_names=[lfk],
                    )
                else:
                    fo_kw.update(
                        no_marshal=True,
                        no_unmarshal=True,
                    )

            if (lon := field.metadata.get(lm.OBJ_MARSHALER_OMIT_IF_NONE)) is not None:
                if check.isinstance(lon, bool):
                    fo_kw.update(
                        omit_if=lang.is_none,
                    )

        if fo_kw.get('embed') and not has_set_name:
            fi_kw.update(
                marshal_name=fi_kw['marshal_name'] + '_',
                unmarshal_names=[n + '_' for n in fi_kw['unmarshal_names']],
            )

        if fo_kw.get('no_marshal'):
            fi_kw.update(
                marshal_name=None,
            )
        if fo_kw.get('no_unmarshal'):
            fi_kw.update(
                unmarshal_names=(),
            )

        ret.append(
            FieldInfo(
                options=FieldOptions(**fo_kw),
                **fi_kw,
            ),
        )

    return FieldInfos(ret)


def _make_field_obj(ctx, ty, obj, fac):
    if obj is not None:
        return obj
    if fac is not None:
        return fac(ctx, ty)
    return ctx.make(ty)


##


class AbstractDataclassFactory(lang.Abstract):
    def _get_metadata(self, ty: type) -> ObjectMetadata:
        return get_dataclass_metadata(ty)

    def _get_field_infos(
            self,
            ty: type,
            opts: col.TypeMap[Option] | None = None,
    ) -> FieldInfos:
        return get_dataclass_field_infos(ty, opts)


##


class DataclassMarshalerFactory(AbstractDataclassFactory, MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and dc.is_dataclass(rty) and not lang.is_abstract_class(rty)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        ty = check.isinstance(rty, type)
        check.state(dc.is_dataclass(ty))
        check.state(not lang.is_abstract_class(ty))

        dc_md = self._get_metadata(ty)
        fis = self._get_field_infos(ty, ctx.options)

        fields = [
            (fi, _make_field_obj(ctx, fi.type, fi.metadata.marshaler, fi.metadata.marshaler_factory))
            for fi in fis
            if fi.name not in dc_md.specials.set
        ]

        return ObjectMarshaler(
            fields,
            specials=dc_md.specials,
        )


##


class DataclassUnmarshalerFactory(AbstractDataclassFactory, UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and dc.is_dataclass(rty) and not lang.is_abstract_class(rty)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        ty = check.isinstance(rty, type)
        check.state(dc.is_dataclass(ty))
        check.state(not lang.is_abstract_class(ty))

        dc_md = self._get_metadata(ty)
        fis = self._get_field_infos(ty, ctx.options)

        d: dict[str, tuple[FieldInfo, Unmarshaler]] = {}
        defaults: dict[str, ta.Any] = {}
        embeds: dict[str, type] = {}
        embeds_by_unmarshal_name: dict[str, tuple[str, str]] = {}

        def add_field(fi: FieldInfo, *, prefixes: ta.Iterable[str] = ('',)) -> ta.Iterable[str]:
            ret: list[str] = []

            if fi.options.embed:
                e_ty = check.isinstance(fi.type, type)
                check.state(dc.is_dataclass(e_ty))
                e_dc_md = get_dataclass_metadata(e_ty)
                if e_dc_md.specials.set:
                    raise Exception(f'Embedded fields cannot have specials: {e_ty}')

                embeds[fi.name] = e_ty
                for e_fi in self._get_field_infos(e_ty, ctx.options):
                    e_ns = add_field(e_fi, prefixes=[p + ep for p in prefixes for ep in fi.unmarshal_names])
                    embeds_by_unmarshal_name.update({e_f: (fi.name, e_fi.name) for e_f in e_ns})
                    ret.extend(e_ns)

            else:
                tup = (fi, _make_field_obj(ctx, fi.type, fi.metadata.unmarshaler, fi.metadata.unmarshaler_factory))

                for pfx in prefixes:
                    for un in fi.unmarshal_names:
                        un = pfx + un
                        if un in d:
                            raise KeyError(f'Duplicate fields for name {un!r}: {fi.name!r}, {d[un][0].name!r}')
                        d[un] = tup
                        ret.append(un)

                if fi.options.default.present:
                    defaults[fi.name] = fi.options.default.must()

            return ret

        for fi in fis:
            if fi.name in dc_md.specials.set:
                continue
            add_field(fi)

        return ObjectUnmarshaler(
            ty,
            d,
            specials=dc_md.specials,
            defaults=defaults,
            embeds=embeds,
            embeds_by_unmarshal_name=embeds_by_unmarshal_name,
            ignore_unknown=dc_md.ignore_unknown,
        )



import collections.abc
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..base import UnmarshalerFactory
from ..values import Value
from .metadata import FieldInfo
from .metadata import FieldInfos
from .metadata import ObjectSpecials


##


@dc.dataclass(frozen=True)
class ObjectUnmarshaler(Unmarshaler):
    factory: ta.Callable
    fields_by_unmarshal_name: ta.Mapping[str, tuple[FieldInfo, Unmarshaler]]

    _: dc.KW_ONLY

    specials: ObjectSpecials = ObjectSpecials()

    defaults: ta.Mapping[str, ta.Any] | None = None

    embeds: ta.Mapping[str, type] | None = None
    embeds_by_unmarshal_name: ta.Mapping[str, tuple[str, str]] | None = None

    ignore_unknown: bool = False

    @classmethod
    def make(
            cls,
            ctx: UnmarshalContext,
            factory: ta.Callable,
            fis: FieldInfos,
            **kwargs: ta.Any,
    ) -> Unmarshaler:
        fields_by_unmarshal_name = {
            n: (fi, ctx.make(fi.type))
            for fi in fis
            for n in fi.unmarshal_names
        }

        defaults = {
            fi.name: fi.options.default.must()
            for fi in fis
            if fi.options.default.present
        }

        return cls(
            factory,
            fields_by_unmarshal_name,
            defaults=defaults,
            **kwargs,
        )

    #

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        ma = check.isinstance(v, collections.abc.Mapping)

        u: ta.Any
        kw: dict[str, ta.Any] = {}
        ukf: dict[str, ta.Any] | None = None

        ekws: dict[str, dict[str, ta.Any]] = {en: {} for en in self.embeds or ()}

        if self.specials.source is not None:
            kw[self.specials.source] = v

        if self.specials.unknown is not None:
            kw[self.specials.unknown] = ukf = {}

        for k, mv in ma.items():
            ks = check.isinstance(k, str)

            try:
                fi, u = self.fields_by_unmarshal_name[ks]

            except KeyError:
                if ukf is not None:
                    ukf[ks] = mv  # FIXME: unmarshal?
                    continue

                if self.ignore_unknown:
                    continue

                raise

            if self.embeds_by_unmarshal_name and (en := self.embeds_by_unmarshal_name.get(ks)):
                tkw, tk = ekws[en[0]], en[1]
            else:
                tkw, tk = kw, fi.name

            if tk in tkw:
                raise KeyError(f'Duplicate keys for field {tk!r}: {ks!r}')

            tkw[tk] = u.unmarshal(ctx, mv)

        for em, ecls in self.embeds.items() if self.embeds else ():
            ekw = ekws[em]
            ev = ecls(**ekw)
            kw[em] = ev

        if self.defaults:
            for dk, dv in self.defaults.items():
                kw.setdefault(dk, dv)

        return self.factory(**kw)


##


@dc.dataclass(frozen=True)
class SimpleObjectUnmarshalerFactory(UnmarshalerFactory):
    dct: ta.Mapping[type, ta.Sequence[FieldInfo]]

    _: dc.KW_ONLY

    specials: ObjectSpecials = ObjectSpecials()

    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and rty in self.dct

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        ty = check.isinstance(rty, type)

        fis = FieldInfos(self.dct[ty])

        return ObjectUnmarshaler.make(
            ctx,
            ty,
            fis,
            specials=self.specials,
        )



"""
TODO:
 - minimal parens
 - text.parts
 - QuoteStyle
 - ParamStyle

==

def needs_parens(self, e: Expr) -> bool:
    if isinstance(e, (Literal, Ident, Name)):
        return True
    elif isinstance(e, Expr):
        return False
    else:
        raise TypeError(e)
"""
import dataclasses as dc
import typing as ta

from ... import dispatch
from ... import lang
from ...text import parts as tp
from ..params import ParamStyle
from ..params import make_params_preparer
from .base import Node
from .binary import Binary
from .binary import BinaryOp
from .binary import BinaryOps
from .exprs import Literal
from .exprs import NameExpr
from .idents import Ident
from .inserts import Insert
from .inserts import Values
from .multi import Multi
from .multi import MultiKind
from .names import Name
from .params import Param
from .relations import Join
from .relations import JoinKind
from .relations import Table
from .selects import Select
from .selects import SelectItem
from .unary import Unary
from .unary import UnaryOp
from .unary import UnaryOps


@dc.dataclass(frozen=True)
class RenderedQueryParts(lang.Final):
    p: tp.Part
    args: lang.Args


@dc.dataclass(frozen=True)
class RenderedQuery(lang.Final):
    s: str
    args: lang.Args


class Renderer(lang.Abstract):
    def __init__(
            self,
            *,
            param_style: ParamStyle | None = None,
    ) -> None:
        super().__init__()

        self._param_style = param_style if param_style is not None else self.default_param_style

        self._params_preparer = make_params_preparer(self._param_style)

    default_param_style: ta.ClassVar[ParamStyle] = ParamStyle.PYFORMAT

    def args(self) -> lang.Args:
        return self._params_preparer.prepare()

    @dispatch.method
    def render(self, o: ta.Any) -> tp.Part:
        raise TypeError(o)

    @classmethod
    def render_query_parts(cls, o: ta.Any, *args: ta.Any, **kwargs: ta.Any) -> RenderedQueryParts:
        r = cls(*args, **kwargs)
        return RenderedQueryParts(
            r.render(o),
            r.args(),
        )

    @classmethod
    def render_query(cls, o: ta.Any, *args: ta.Any, **kwargs: ta.Any) -> RenderedQuery:
        rqp = cls.render_query_parts(o, *args, **kwargs)
        return RenderedQuery(
            tp.render(rqp.p),
            rqp.args,
        )


class StdRenderer(Renderer):
    # parens

    NEEDS_PAREN_TYPES: ta.AbstractSet[type[Node]] = {
        Binary,
        # IsNull,
        # SelectExpr,
    }

    def needs_paren(self, node: Node) -> bool:
        return type(node) in self.NEEDS_PAREN_TYPES

    def paren(self, node: Node) -> tp.Part:
        return tp.Wrap(self.render(node)) if self.needs_paren(node) else self.render(node)

    # binary

    BINARY_OP_TO_STR: ta.ClassVar[ta.Mapping[BinaryOp, str]] = {
        BinaryOps.EQ: '=',
        BinaryOps.NE: '!=',
        BinaryOps.LT: '<',
        BinaryOps.LE: '<=',
        BinaryOps.GT: '>',
        BinaryOps.GE: '>=',

        BinaryOps.ADD: '+',
        BinaryOps.SUB: '-',
        BinaryOps.MUL: '*',
        BinaryOps.DIV: '/',
        BinaryOps.MOD: '%',

        BinaryOps.CONCAT: '||',
    }

    @Renderer.render.register
    def render_binary(self, o: Binary) -> tp.Part:
        return [
            self.paren(o.l),
            self.BINARY_OP_TO_STR[o.op],
            self.paren(o.r),
        ]

    # exprs

    @Renderer.render.register
    def render_literal(self, o: Literal) -> tp.Part:
        return repr(o.v)

    @Renderer.render.register
    def render_name_expr(self, o: NameExpr) -> tp.Part:
        return self.render(o.n)

    @Renderer.render.register
    def render_param(self, o: Param) -> tp.Part:
        return self._params_preparer.add(o.n if o.n is not None else id(o))

    # idents

    @Renderer.render.register
    def render_ident(self, o: Ident) -> tp.Part:
        return f'"{o.s}"'

    # inserts

    @Renderer.render.register
    def render_values(self, o: Values) -> tp.Part:
        return [
            'values',
            tp.Wrap(tp.List([self.render(v) for v in o.vs])),
        ]

    @Renderer.render.register
    def render_insert(self, o: Insert) -> tp.Part:
        return [
            'insert into',
            self.render(o.into),
            tp.Wrap(tp.List([self.render(c) for c in o.columns])),
            self.render(o.data),
        ]

    # multis

    MULTI_KIND_TO_STR: ta.ClassVar[ta.Mapping[MultiKind, str]] = {
        MultiKind.AND: 'and',
        MultiKind.OR: 'or',
    }

    @Renderer.render.register
    def render_multi(self, o: Multi) -> tp.Part:
        return tp.Wrap(tp.List(
            [self.render(e) for e in o.es],
            delimiter=' ' + self.MULTI_KIND_TO_STR[o.k],  # FIXME: Part
        ))

    # names

    @Renderer.render.register
    def render_name(self, o: Name) -> tp.Part:
        out: list[tp.Part] = []
        for n, i in enumerate(o.ps):
            if n:
                out.append('.')
            out.append(self.render(i))
        return tp.Concat(out)

    # relations

    @Renderer.render.register
    def render_table(self, o: Table) -> tp.Part:
        return [
            self.render(o.n),
            *(['as', self.render(o.a)] if o.a is not None else []),
        ]

    JOIN_KIND_TO_STR: ta.ClassVar[ta.Mapping[JoinKind, str]] = {
        JoinKind.DEFAULT: 'join',
        JoinKind.INNER: 'inner join',
        JoinKind.LEFT: 'left join',
        JoinKind.LEFT_OUTER: 'left outer join',
        JoinKind.RIGHT: 'right join',
        JoinKind.RIGHT_OUTER: 'right outer join',
        JoinKind.FULL: 'full join',
        JoinKind.FULL_OUTER: 'full outer join',
        JoinKind.CROSS: 'cross join',
        JoinKind.NATURAL: 'natural join',
    }

    @Renderer.render.register
    def render_join(self, o: Join) -> tp.Part:
        return [
            self.render(o.l),
            self.JOIN_KIND_TO_STR[o.k],
            self.render(o.r),
            *(['on', self.render(o.c)] if o.c is not None else []),
        ]

    # selects

    @Renderer.render.register
    def render_select_item(self, o: SelectItem) -> tp.Part:
        return [
            self.render(o.v),
            *(['as', self.render(o.a)] if o.a is not None else []),
        ]

    @Renderer.render.register
    def render_select(self, o: Select) -> tp.Part:
        return [
            'select',
            tp.List([self.render(i) for i in o.items]),
            *(['from', self.render(o.from_)] if o.from_ is not None else []),
            *(['where', self.render(o.where)] if o.where is not None else []),
        ]

    # unary

    UNARY_OP_TO_STR: ta.ClassVar[ta.Mapping[UnaryOp, tuple[str, str]]] = {
        UnaryOps.NOT: ('not ', ''),
        UnaryOps.IS_NULL: ('', ' is null'),
        UnaryOps.IS_NOT_NULL: ('', ' is not null'),

        UnaryOps.POS: ('+', ''),
        UnaryOps.NEG: ('-', ''),
    }

    @Renderer.render.register
    def render_unary(self, o: Unary) -> tp.Part:
        pfx, sfx = self.UNARY_OP_TO_STR[o.op]
        return tp.Concat([
            pfx,
            self.render(o.v),
            sfx,
        ])


def render_parts(n: Node, **kwargs: ta.Any) -> RenderedQueryParts:
    return StdRenderer.render_query_parts(n, **kwargs)


def render(n: Node, **kwargs: ta.Any) -> RenderedQuery:
    return StdRenderer.render_query(n, **kwargs)



"""
TODO:
 - Rewrite lol
 - Enum - enforce Abstract or Final
"""
import abc
import collections
import dataclasses as dc
import typing as ta

from ... import lang
from .api import MISSING
from .api import dataclass
from .api import field  # noqa
from .params import MetaclassParams
from .params import get_metaclass_params
from .params import get_params
from .params import get_params_extras


T = ta.TypeVar('T')


##


_CONFER_PARAMS: tuple[str, ...] = (
    'frozen',
    'kw_only',
)

_CONFER_PARAMS_EXTRAS: tuple[str, ...] = (
    'reorder',
    'cache_hash',
    'generic_init',
    'override',
)

_CONFER_METACLASS_PARAMS: tuple[str, ...] = (
    'confer',
    'final_subclasses',
    'abstract_immediate_subclasses',
)


def confer_kwarg(out: dict[str, ta.Any], k: str, v: ta.Any) -> None:
    if k in out:
        if out[k] != v:
            raise ValueError
    else:
        out[k] = v


def confer_kwargs(
        bases: ta.Sequence[type],
        kwargs: ta.Mapping[str, ta.Any],
) -> dict[str, ta.Any]:
    out: dict[str, ta.Any] = {}
    for base in bases:
        if not dc.is_dataclass(base):
            continue

        if not (bmp := get_metaclass_params(base)).confer:
            continue

        for ck in bmp.confer:
            if ck in kwargs:
                continue

            if ck in _CONFER_PARAMS:
                confer_kwarg(out, ck, getattr(get_params(base), ck))

            elif ck in _CONFER_PARAMS_EXTRAS:
                confer_kwarg(out, ck, getattr(get_params_extras(base), ck))

            elif ck in _CONFER_METACLASS_PARAMS:
                confer_kwarg(out, ck, getattr(bmp, ck))

            else:
                raise KeyError(ck)

    return out


##


class DataMeta(abc.ABCMeta):
    def __new__(
            mcls,
            name,
            bases,
            namespace,
            *,

            abstract=False,
            sealed=False,
            final=False,

            metadata=None,
            **kwargs,
    ):
        ckw = confer_kwargs(bases, kwargs)
        nkw = {**kwargs, **ckw}

        mcp = MetaclassParams(**{
            mpa: nkw.pop(mpa)
            for mpa in _CONFER_METACLASS_PARAMS
            if mpa in nkw
        })

        mmd = {
            MetaclassParams: mcp,
        }
        if metadata is not None:
            metadata = collections.ChainMap(mmd, metadata)
        else:
            metadata = mmd

        #

        xbs: list[type] = []

        if any(get_metaclass_params(b).abstract_immediate_subclasses for b in bases if dc.is_dataclass(b)):
            abstract = True

        final |= (mcp.final_subclasses and not abstract)

        if final and abstract:
            raise TypeError(f'Class cannot be abstract and final: {name!r}')

        if abstract:
            xbs.append(lang.Abstract)
        if sealed:
            xbs.append(lang.Sealed)
        if final:
            xbs.append(lang.Final)

        if xbs:
            if bases and bases[-1] is ta.Generic:
                bases = (*bases[:-1], *xbs, bases[-1])
            else:
                bases = (*bases, *xbs)
            if ob := namespace.get('__orig_bases__'):
                if getattr(ob[-1], '__origin__', None) is ta.Generic:
                    namespace['__orig_bases__'] = (*ob[:-1], *xbs, ob[-1])
                else:
                    namespace['__orig_bases__'] = (*ob, *xbs)

        #

        ofs: set[str] = set()
        if any(issubclass(b, lang.Abstract) for b in bases) and nkw.get('override'):
            ofs.update(a for a in namespace.get('__annotations__', []) if a not in namespace)
            namespace.update((a, MISSING) for a in ofs)

        #

        cls = lang.super_meta(
            super(),
            mcls,
            name,
            bases,
            namespace,
        )

        #

        for a in ofs:
            delattr(cls, a)

        #

        return dataclass(cls, metadata=metadata, **nkw)


##


# @ta.dataclass_transform(field_specifiers=(field,))  # FIXME: ctor
class Data(
    eq=False,
    order=False,
    confer=frozenset([
        'confer',
        'final_subclasses',
    ]),
    metaclass=DataMeta,
):
    def __init__(self, *args, **kwargs):
        # Typechecking barrier
        super().__init__(*args, **kwargs)

    def __init_subclass__(cls, **kwargs):
        # Typechecking barrier
        super().__init_subclass__(**kwargs)

    def __post_init__(self, *args, **kwargs) -> None:
        try:
            spi = super().__post_init__  # type: ignore  # noqa
        except AttributeError:
            if args or kwargs:
                raise TypeError(args, kwargs) from None
        else:
            spi(*args, **kwargs)


class Frozen(
    Data,
    frozen=True,
    eq=False,
    order=False,
    confer=frozenset([
        *get_metaclass_params(Data).confer,
        'frozen',
        'reorder',
        'cache_hash',
        'override',
    ]),
):
    pass


class Case(
    Frozen,
    abstract=True,
    override=True,
    final_subclasses=True,
    abstract_immediate_subclasses=True,
):
    pass


class Box(
    Frozen,
    ta.Generic[T],
    generic_init=True,
    confer=frozenset([
        *get_metaclass_params(Frozen).confer,
        'generic_init',
    ]),
):
    v: T



from .base import (  # noqa
    EagerCodec,
    IncrementalCodec,
    ComboCodec,

    check_codec_name,

    Codec,

    LazyLoadedCodec,
)

from .bytes import (  # noqa
    ASCII85,
    BASE16,
    BASE32,
    BASE64,
    BASE85,
    BASE32_HEX,
    BASE64_HEX,
    BASE64_URLSAFE,
    HEX,
)

from .chain import (  # noqa
    ChainEagerCodec,

    chain,
)

from .funcs import (  # noqa
    FnPairEagerCodec,

    of_pair,
    of,
)

from .registry import (  # noqa
    CodecRegistry,

    REGISTRY,
    register,
    lookup,

    encode,
    decode,
)

from .standard import (  # noqa
    STANDARD_CODECS,
)

from .text import (  # noqa
    TextEncodingErrors,
    TextEncodingOptions,

    TextEncodingComboCodec,

    TextEncodingCodec,

    ASCII,
    LATIN1,
    UTF32,
    UTF32BE,
    UTF32LE,
    UTF16,
    UTF16BE,
    UTF16LE,
    UTF7,
    UTF8,
    UTF8SIG,
)



# ruff: noqa: UP006 UP007
# @omlish-lite
import typing as ta

from ...sockets.addresses import SocketAddress
from ...sockets.handlers import SocketHandler_
from ...sockets.io import SocketIoPair
from .server import CoroHttpServer
from .server import CoroHttpServerFactory


##


class CoroHttpServerSocketHandler(SocketHandler_):
    def __init__(
            self,
            server_factory: CoroHttpServerFactory,
            *,
            log_handler: ta.Optional[ta.Callable[[CoroHttpServer, CoroHttpServer.AnyLogIo], None]] = None,
    ) -> None:
        super().__init__()

        self._server_factory = server_factory
        self._log_handler = log_handler

    def __call__(self, client_address: SocketAddress, fp: SocketIoPair) -> None:
        server = self._server_factory(client_address)

        gen = server.coro_handle()

        o = next(gen)
        while True:
            if isinstance(o, CoroHttpServer.AnyLogIo):
                i = None
                if self._log_handler is not None:
                    self._log_handler(server, o)

            elif isinstance(o, CoroHttpServer.ReadIo):
                i = fp.r.read(o.sz)

            elif isinstance(o, CoroHttpServer.ReadLineIo):
                i = fp.r.readline(o.sz)

            elif isinstance(o, CoroHttpServer.WriteIo):
                i = None
                fp.w.write(o.data)
                fp.w.flush()

            else:
                raise TypeError(o)

            try:
                if i is not None:
                    o = gen.send(i)
                else:
                    o = next(gen)
            except StopIteration:
                break



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#/

# When we hit an accept state in either the DFA or the ATN, we
#  have to notify the character stream to start buffering characters
#  via {@link IntStream#mark} and record the current state. The current sim state
#  includes the current index into the input, the current line,
#  and current character position in that line. Note that the Lexer is
#  tracking the starting line and characterization of the token. These
#  variables track the "state" of the simulator when it hits an accept state.
#
#  <p>We track these variables separately for the DFA and ATN simulation
#  because the DFA simulation often has to fail over to the ATN
#  simulation. If the ATN simulation fails, we need the DFA to fall
#  back to its previously accepted state, if any. If the ATN succeeds,
#  then the ATN does the accept and the DFA simulator that invoked it
#  can simply return the predicted token type.</p>
#/

from ..PredictionContext import PredictionContextCache, SingletonPredictionContext, PredictionContext
from ..InputStream import InputStream
from ..Token import Token
from .ATN import ATN
from .ATNConfig import LexerATNConfig
from .ATNSimulator import ATNSimulator
from .ATNConfigSet import ATNConfigSet, OrderedATNConfigSet
from .ATNState import RuleStopState, ATNState
from .LexerActionExecutor import LexerActionExecutor
from .Transition import Transition
from ..dfa.DFAState import DFAState
from ..error.Errors import LexerNoViableAltException, UnsupportedOperationException

class SimState(object):
    __slots__ = ('index', 'line', 'column', 'dfaState')

    def __init__(self):
        self.reset()

    def reset(self):
        self.index = -1
        self.line = 0
        self.column = -1
        self.dfaState = None

# need forward declaration
Lexer = None
LexerATNSimulator = None

class LexerATNSimulator(ATNSimulator):
    __slots__ = (
        'decisionToDFA', 'recog', 'startIndex', 'line', 'column', 'mode',
        'DEFAULT_MODE', 'MAX_CHAR_VALUE', 'prevAccept'
    )

    debug = False
    dfa_debug = False

    MIN_DFA_EDGE = 0
    MAX_DFA_EDGE = 127 # forces unicode to stay in ATN

    ERROR = None

    def __init__(self, recog:Lexer, atn:ATN, decisionToDFA:list, sharedContextCache:PredictionContextCache):
        super().__init__(atn, sharedContextCache)
        self.decisionToDFA = decisionToDFA
        self.recog = recog
        # The current token's starting index into the character stream.
        #  Shared across DFA to ATN simulation in case the ATN fails and the
        #  DFA did not have a previous accept state. In this case, we use the
        #  ATN-generated exception object.
        self.startIndex = -1
        # line number 1..n within the input#/
        self.line = 1
        # The index of the character relative to the beginning of the line 0..n-1#/
        self.column = 0
        from ..Lexer import Lexer
        self.mode = Lexer.DEFAULT_MODE
        # Cache Lexer properties to avoid further imports
        self.DEFAULT_MODE = Lexer.DEFAULT_MODE
        self.MAX_CHAR_VALUE = Lexer.MAX_CHAR_VALUE
        # Used during DFA/ATN exec to record the most recent accept configuration info
        self.prevAccept = SimState()


    def copyState(self, simulator:LexerATNSimulator ):
        self.column = simulator.column
        self.line = simulator.line
        self.mode = simulator.mode
        self.startIndex = simulator.startIndex

    def match(self, input:InputStream , mode:int):
        self.mode = mode
        mark = input.mark()
        try:
            self.startIndex = input.index
            self.prevAccept.reset()
            dfa = self.decisionToDFA[mode]
            if dfa.s0 is None:
                return self.matchATN(input)
            else:
                return self.execATN(input, dfa.s0)
        finally:
            input.release(mark)

    def reset(self):
        self.prevAccept.reset()
        self.startIndex = -1
        self.line = 1
        self.column = 0
        self.mode = self.DEFAULT_MODE

    def matchATN(self, input:InputStream):
        startState = self.atn.modeToStartState[self.mode]

        if LexerATNSimulator.debug:
            print("matchATN mode " + str(self.mode) + " start: " + str(startState))

        old_mode = self.mode
        s0_closure = self.computeStartState(input, startState)
        suppressEdge = s0_closure.hasSemanticContext
        s0_closure.hasSemanticContext = False

        next = self.addDFAState(s0_closure)
        if not suppressEdge:
            self.decisionToDFA[self.mode].s0 = next

        predict = self.execATN(input, next)

        if LexerATNSimulator.debug:
            print("DFA after matchATN: " + str(self.decisionToDFA[old_mode].toLexerString()))

        return predict

    def execATN(self, input:InputStream, ds0:DFAState):
        if LexerATNSimulator.debug:
            print("start state closure=" + str(ds0.configs))

        if ds0.isAcceptState:
            # allow zero-length tokens
            self.captureSimState(self.prevAccept, input, ds0)

        t = input.LA(1)
        s = ds0 # s is current/from DFA state

        while True: # while more work
            if LexerATNSimulator.debug:
                print("execATN loop starting closure:", str(s.configs))

            # As we move src->trg, src->trg, we keep track of the previous trg to
            # avoid looking up the DFA state again, which is expensive.
            # If the previous target was already part of the DFA, we might
            # be able to avoid doing a reach operation upon t. If s!=null,
            # it means that semantic predicates didn't prevent us from
            # creating a DFA state. Once we know s!=null, we check to see if
            # the DFA state has an edge already for t. If so, we can just reuse
            # it's configuration set; there's no point in re-computing it.
            # This is kind of like doing DFA simulation within the ATN
            # simulation because DFA simulation is really just a way to avoid
            # computing reach/closure sets. Technically, once we know that
            # we have a previously added DFA state, we could jump over to
            # the DFA simulator. But, that would mean popping back and forth
            # a lot and making things more complicated algorithmically.
            # This optimization makes a lot of sense for loops within DFA.
            # A character will take us back to an existing DFA state
            # that already has lots of edges out of it. e.g., .* in comments.
            # print("Target for:" + str(s) + " and:" + str(t))
            target = self.getExistingTargetState(s, t)
            # print("Existing:" + str(target))
            if target is None:
                target = self.computeTargetState(input, s, t)
                # print("Computed:" + str(target))

            if target == self.ERROR:
                break

            # If this is a consumable input element, make sure to consume before
            # capturing the accept state so the input index, line, and char
            # position accurately reflect the state of the interpreter at the
            # end of the token.
            if t != Token.EOF:
                self.consume(input)

            if target.isAcceptState:
                self.captureSimState(self.prevAccept, input, target)
                if t == Token.EOF:
                    break

            t = input.LA(1)

            s = target # flip; current DFA target becomes new src/from state

        return self.failOrAccept(self.prevAccept, input, s.configs, t)

    # Get an existing target state for an edge in the DFA. If the target state
    # for the edge has not yet been computed or is otherwise not available,
    # this method returns {@code null}.
    #
    # @param s The current DFA state
    # @param t The next input symbol
    # @return The existing target DFA state for the given input symbol
    # {@code t}, or {@code null} if the target state for this edge is not
    # already cached
    def getExistingTargetState(self, s:DFAState, t:int):
        if s.edges is None or t < self.MIN_DFA_EDGE or t > self.MAX_DFA_EDGE:
            return None

        target = s.edges[t - self.MIN_DFA_EDGE]
        if LexerATNSimulator.debug and target is not None:
            print("reuse state", str(s.stateNumber), "edge to", str(target.stateNumber))

        return target

    # Compute a target state for an edge in the DFA, and attempt to add the
    # computed state and corresponding edge to the DFA.
    #
    # @param input The input stream
    # @param s The current DFA state
    # @param t The next input symbol
    #
    # @return The computed target DFA state for the given input symbol
    # {@code t}. If {@code t} does not lead to a valid DFA state, this method
    # returns {@link #ERROR}.
    def computeTargetState(self, input:InputStream, s:DFAState, t:int):
        reach = OrderedATNConfigSet()

        # if we don't find an existing DFA state
        # Fill reach starting from closure, following t transitions
        self.getReachableConfigSet(input, s.configs, reach, t)

        if len(reach)==0: # we got nowhere on t from s
            if not reach.hasSemanticContext:
                # we got nowhere on t, don't throw out this knowledge; it'd
                # cause a failover from DFA later.
               self. addDFAEdge(s, t, self.ERROR)

            # stop when we can't match any more char
            return self.ERROR

        # Add an edge from s to target DFA found/created for reach
        return self.addDFAEdge(s, t, cfgs=reach)

    def failOrAccept(self, prevAccept:SimState , input:InputStream, reach:ATNConfigSet, t:int):
        if self.prevAccept.dfaState is not None:
            lexerActionExecutor = prevAccept.dfaState.lexerActionExecutor
            self.accept(input, lexerActionExecutor, self.startIndex, prevAccept.index, prevAccept.line, prevAccept.column)
            return prevAccept.dfaState.prediction
        else:
            # if no accept and EOF is first char, return EOF
            if t==Token.EOF and input.index==self.startIndex:
                return Token.EOF
            raise LexerNoViableAltException(self.recog, input, self.startIndex, reach)

    # Given a starting configuration set, figure out all ATN configurations
    #  we can reach upon input {@code t}. Parameter {@code reach} is a return
    #  parameter.
    def getReachableConfigSet(self, input:InputStream, closure:ATNConfigSet, reach:ATNConfigSet, t:int):
        # this is used to skip processing for configs which have a lower priority
        # than a config that already reached an accept state for the same rule
        skipAlt = ATN.INVALID_ALT_NUMBER
        for cfg in closure:
            currentAltReachedAcceptState = ( cfg.alt == skipAlt )
            if currentAltReachedAcceptState and cfg.passedThroughNonGreedyDecision:
                continue

            if LexerATNSimulator.debug:
                print("testing", self.getTokenName(t), "at",  str(cfg))

            for trans in cfg.state.transitions:          # for each transition
                target = self.getReachableTarget(trans, t)
                if target is not None:
                    lexerActionExecutor = cfg.lexerActionExecutor
                    if lexerActionExecutor is not None:
                        lexerActionExecutor = lexerActionExecutor.fixOffsetBeforeMatch(input.index - self.startIndex)

                    treatEofAsEpsilon = (t == Token.EOF)
                    config = LexerATNConfig(state=target, lexerActionExecutor=lexerActionExecutor, config=cfg)
                    if self.closure(input, config, reach, currentAltReachedAcceptState, True, treatEofAsEpsilon):
                        # any remaining configs for this alt have a lower priority than
                        # the one that just reached an accept state.
                        skipAlt = cfg.alt

    def accept(self, input:InputStream, lexerActionExecutor:LexerActionExecutor, startIndex:int, index:int, line:int, charPos:int):
        if LexerATNSimulator.debug:
            print("ACTION", lexerActionExecutor)

        # seek to after last char in token
        input.seek(index)
        self.line = line
        self.column = charPos

        if lexerActionExecutor is not None and self.recog is not None:
            lexerActionExecutor.execute(self.recog, input, startIndex)

    def getReachableTarget(self, trans:Transition, t:int):
        if trans.matches(t, 0, self.MAX_CHAR_VALUE):
            return trans.target
        else:
            return None

    def computeStartState(self, input:InputStream, p:ATNState):
        initialContext = PredictionContext.EMPTY
        configs = OrderedATNConfigSet()
        for i in range(0,len(p.transitions)):
            target = p.transitions[i].target
            c = LexerATNConfig(state=target, alt=i+1, context=initialContext)
            self.closure(input, c, configs, False, False, False)
        return configs

    # Since the alternatives within any lexer decision are ordered by
    # preference, this method stops pursuing the closure as soon as an accept
    # state is reached. After the first accept state is reached by depth-first
    # search from {@code config}, all other (potentially reachable) states for
    # this rule would have a lower priority.
    #
    # @return {@code true} if an accept state is reached, otherwise
    # {@code false}.
    def closure(self, input:InputStream, config:LexerATNConfig, configs:ATNConfigSet, currentAltReachedAcceptState:bool,
                speculative:bool, treatEofAsEpsilon:bool):
        if LexerATNSimulator.debug:
            print("closure(" + str(config) + ")")

        if isinstance( config.state, RuleStopState ):
            if LexerATNSimulator.debug:
                if self.recog is not None:
                    print("closure at", self.recog.symbolicNames[config.state.ruleIndex],  "rule stop", str(config))
                else:
                    print("closure at rule stop", str(config))

            if config.context is None or config.context.hasEmptyPath():
                if config.context is None or config.context.isEmpty():
                    configs.add(config)
                    return True
                else:
                    configs.add(LexerATNConfig(state=config.state, config=config, context=PredictionContext.EMPTY))
                    currentAltReachedAcceptState = True

            if config.context is not None and not config.context.isEmpty():
                for i in range(0,len(config.context)):
                    if config.context.getReturnState(i) != PredictionContext.EMPTY_RETURN_STATE:
                        newContext = config.context.getParent(i) # "pop" return state
                        returnState = self.atn.states[config.context.getReturnState(i)]
                        c = LexerATNConfig(state=returnState, config=config, context=newContext)
                        currentAltReachedAcceptState = self.closure(input, c, configs,
                                    currentAltReachedAcceptState, speculative, treatEofAsEpsilon)

            return currentAltReachedAcceptState

        # optimization
        if not config.state.epsilonOnlyTransitions:
            if not currentAltReachedAcceptState or not config.passedThroughNonGreedyDecision:
                configs.add(config)

        for t in config.state.transitions:
            c = self.getEpsilonTarget(input, config, t, configs, speculative, treatEofAsEpsilon)
            if c is not None:
                currentAltReachedAcceptState = self.closure(input, c, configs, currentAltReachedAcceptState, speculative, treatEofAsEpsilon)

        return currentAltReachedAcceptState

    # side-effect: can alter configs.hasSemanticContext
    def getEpsilonTarget(self, input:InputStream, config:LexerATNConfig, t:Transition, configs:ATNConfigSet,
                                           speculative:bool, treatEofAsEpsilon:bool):
        c = None
        if t.serializationType==Transition.RULE:
                newContext = SingletonPredictionContext.create(config.context, t.followState.stateNumber)
                c = LexerATNConfig(state=t.target, config=config, context=newContext)

        elif t.serializationType==Transition.PRECEDENCE:
                raise UnsupportedOperationException("Precedence predicates are not supported in lexers.")

        elif t.serializationType==Transition.PREDICATE:
                #  Track traversing semantic predicates. If we traverse,
                # we cannot add a DFA state for this "reach" computation
                # because the DFA would not test the predicate again in the
                # future. Rather than creating collections of semantic predicates
                # like v3 and testing them on prediction, v4 will test them on the
                # fly all the time using the ATN not the DFA. This is slower but
                # semantically it's not used that often. One of the key elements to
                # this predicate mechanism is not adding DFA states that see
                # predicates immediately afterwards in the ATN. For example,

                # a : ID {p1}? | ID {p2}? ;

                # should create the start state for rule 'a' (to save start state
                # competition), but should not create target of ID state. The
                # collection of ATN states the following ID references includes
                # states reached by traversing predicates. Since this is when we
                # test them, we cannot cash the DFA state target of ID.

                if LexerATNSimulator.debug:
                    print("EVAL rule "+ str(t.ruleIndex) + ":" + str(t.predIndex))
                configs.hasSemanticContext = True
                if self.evaluatePredicate(input, t.ruleIndex, t.predIndex, speculative):
                    c = LexerATNConfig(state=t.target, config=config)

        elif t.serializationType==Transition.ACTION:
                if config.context is None or config.context.hasEmptyPath():
                    # execute actions anywhere in the start rule for a token.
                    #
                    # TODO: if the entry rule is invoked recursively, some
                    # actions may be executed during the recursive call. The
                    # problem can appear when hasEmptyPath() is true but
                    # isEmpty() is false. In this case, the config needs to be
                    # split into two contexts - one with just the empty path
                    # and another with everything but the empty path.
                    # Unfortunately, the current algorithm does not allow
                    # getEpsilonTarget to return two configurations, so
                    # additional modifications are needed before we can support
                    # the split operation.
                    lexerActionExecutor = LexerActionExecutor.append(config.lexerActionExecutor,
                                    self.atn.lexerActions[t.actionIndex])
                    c = LexerATNConfig(state=t.target, config=config, lexerActionExecutor=lexerActionExecutor)

                else:
                    # ignore actions in referenced rules
                    c = LexerATNConfig(state=t.target, config=config)

        elif t.serializationType==Transition.EPSILON:
            c = LexerATNConfig(state=t.target, config=config)

        elif t.serializationType in [ Transition.ATOM, Transition.RANGE, Transition.SET ]:
            if treatEofAsEpsilon:
                if t.matches(Token.EOF, 0, self.MAX_CHAR_VALUE):
                    c = LexerATNConfig(state=t.target, config=config)

        return c

    # Evaluate a predicate specified in the lexer.
    #
    # <p>If {@code speculative} is {@code true}, this method was called before
    # {@link #consume} for the matched character. This method should call
    # {@link #consume} before evaluating the predicate to ensure position
    # sensitive values, including {@link Lexer#getText}, {@link Lexer#getLine},
    # and {@link Lexer#getcolumn}, properly reflect the current
    # lexer state. This method should restore {@code input} and the simulator
    # to the original state before returning (i.e. undo the actions made by the
    # call to {@link #consume}.</p>
    #
    # @param input The input stream.
    # @param ruleIndex The rule containing the predicate.
    # @param predIndex The index of the predicate within the rule.
    # @param speculative {@code true} if the current index in {@code input} is
    # one character before the predicate's location.
    #
    # @return {@code true} if the specified predicate evaluates to
    # {@code true}.
    #/
    def evaluatePredicate(self, input:InputStream, ruleIndex:int, predIndex:int, speculative:bool):
        # assume true if no recognizer was provided
        if self.recog is None:
            return True

        if not speculative:
            return self.recog.sempred(None, ruleIndex, predIndex)

        savedcolumn = self.column
        savedLine = self.line
        index = input.index
        marker = input.mark()
        try:
            self.consume(input)
            return self.recog.sempred(None, ruleIndex, predIndex)
        finally:
            self.column = savedcolumn
            self.line = savedLine
            input.seek(index)
            input.release(marker)

    def captureSimState(self, settings:SimState, input:InputStream, dfaState:DFAState):
        settings.index = input.index
        settings.line = self.line
        settings.column = self.column
        settings.dfaState = dfaState

    def addDFAEdge(self, from_:DFAState, tk:int, to:DFAState=None, cfgs:ATNConfigSet=None) -> DFAState:

        if to is None and cfgs is not None:
            # leading to this call, ATNConfigSet.hasSemanticContext is used as a
            # marker indicating dynamic predicate evaluation makes this edge
            # dependent on the specific input sequence, so the static edge in the
            # DFA should be omitted. The target DFAState is still created since
            # execATN has the ability to resynchronize with the DFA state cache
            # following the predicate evaluation step.
            #
            # TJP notes: next time through the DFA, we see a pred again and eval.
            # If that gets us to a previously created (but dangling) DFA
            # state, we can continue in pure DFA mode from there.
            #/
            suppressEdge = cfgs.hasSemanticContext
            cfgs.hasSemanticContext = False

            to = self.addDFAState(cfgs)

            if suppressEdge:
                return to

        # add the edge
        if tk < self.MIN_DFA_EDGE or tk > self.MAX_DFA_EDGE:
            # Only track edges within the DFA bounds
            return to

        if LexerATNSimulator.debug:
            print("EDGE " + str(from_) + " -> " + str(to) + " upon "+ chr(tk))

        if from_.edges is None:
            #  make room for tokens 1..n and -1 masquerading as index 0
            from_.edges = [ None ] * (self.MAX_DFA_EDGE - self.MIN_DFA_EDGE + 1)

        from_.edges[tk - self.MIN_DFA_EDGE] = to # connect

        return to


    # Add a new DFA state if there isn't one with this set of
    # configurations already. This method also detects the first
    # configuration containing an ATN rule stop state. Later, when
    # traversing the DFA, we will know which rule to accept.
    def addDFAState(self, configs:ATNConfigSet) -> DFAState:

        proposed = DFAState(configs=configs)
        firstConfigWithRuleStopState = next((cfg for cfg in configs if isinstance(cfg.state, RuleStopState)), None)

        if firstConfigWithRuleStopState is not None:
            proposed.isAcceptState = True
            proposed.lexerActionExecutor = firstConfigWithRuleStopState.lexerActionExecutor
            proposed.prediction = self.atn.ruleToTokenType[firstConfigWithRuleStopState.state.ruleIndex]

        dfa = self.decisionToDFA[self.mode]
        existing = dfa.states.get(proposed, None)
        if existing is not None:
            return existing

        newState = proposed

        newState.stateNumber = len(dfa.states)
        configs.setReadonly(True)
        newState.configs = configs
        dfa.states[newState] = newState
        return newState

    def getDFA(self, mode:int):
        return self.decisionToDFA[mode]

    # Get the text matched so far for the current token.
    def getText(self, input:InputStream):
        # index is first lookahead char, don't include.
        return input.getText(self.startIndex, input.index-1)

    def consume(self, input:InputStream):
        curChar = input.LA(1)
        if curChar==ord('\n'):
            self.line += 1
            self.column = 0
        else:
            self.column += 1
        input.consume()

    def getTokenName(self, t:int):
        if t==-1:
            return "EOF"
        else:
            return "'" + chr(t) + "'"


LexerATNSimulator.ERROR = DFAState(0x7FFFFFFF, ATNConfigSet())

del Lexer
